from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
    TimeSeries,
)
from ._price_area import PriceArea, PriceAreaWrite

if TYPE_CHECKING:
    from ._bid_configuration_day_ahead import (
        BidConfigurationDayAhead,
        BidConfigurationDayAheadGraphQL,
        BidConfigurationDayAheadWrite,
    )


__all__ = [
    "PriceAreaDayAhead",
    "PriceAreaDayAheadWrite",
    "PriceAreaDayAheadApply",
    "PriceAreaDayAheadList",
    "PriceAreaDayAheadWriteList",
    "PriceAreaDayAheadApplyList",
    "PriceAreaDayAheadFields",
    "PriceAreaDayAheadTextFields",
]


PriceAreaDayAheadTextFields = Literal["name", "display_name", "asset_type", "main_price_scenario", "price_scenarios"]
PriceAreaDayAheadFields = Literal[
    "name", "display_name", "ordering", "asset_type", "main_price_scenario", "price_scenarios"
]

_PRICEAREADAYAHEAD_PROPERTIES_BY_FIELD = {
    "name": "name",
    "display_name": "displayName",
    "ordering": "ordering",
    "asset_type": "assetType",
    "main_price_scenario": "mainPriceScenario",
    "price_scenarios": "priceScenarios",
}


class PriceAreaDayAheadGraphQL(GraphQLCore):
    """This represents the reading version of price area day ahead, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area day ahead.
        data_record: The data record of the price area day ahead node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        default_bid_configuration: TODO
        main_price_scenario: TODO
        price_scenarios: TODO
    """

    view_id = dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1")
    name: Optional[str] = None
    display_name: Optional[str] = Field(None, alias="displayName")
    ordering: Optional[int] = None
    asset_type: Optional[str] = Field(None, alias="assetType")
    default_bid_configuration: Optional[BidConfigurationDayAheadGraphQL] = Field(
        default=None, repr=False, alias="defaultBidConfiguration"
    )
    main_price_scenario: Union[TimeSeries, dict, None] = Field(None, alias="mainPriceScenario")
    price_scenarios: Union[list[TimeSeries], list[dict], None] = Field(None, alias="priceScenarios")

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    @field_validator("price_scenarios", mode="before")
    def clean_list(cls, value: Any) -> Any:
        if isinstance(value, list):
            return [v for v in value if v is not None] or None
        return value

    @field_validator("default_bid_configuration", mode="before")
    def parse_graphql(cls, value: Any) -> Any:
        if not isinstance(value, dict):
            return value
        if "items" in value:
            return value["items"]
        return value

    def as_read(self) -> PriceAreaDayAhead:
        """Convert this GraphQL format of price area day ahead to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return PriceAreaDayAhead(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
            asset_type=self.asset_type,
            default_bid_configuration=(
                self.default_bid_configuration.as_read()
                if isinstance(self.default_bid_configuration, GraphQLCore)
                else self.default_bid_configuration
            ),
            main_price_scenario=self.main_price_scenario,
            price_scenarios=self.price_scenarios,
        )

    def as_write(self) -> PriceAreaDayAheadWrite:
        """Convert this GraphQL format of price area day ahead to the writing format."""
        return PriceAreaDayAheadWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
            asset_type=self.asset_type,
            default_bid_configuration=(
                self.default_bid_configuration.as_write()
                if isinstance(self.default_bid_configuration, GraphQLCore)
                else self.default_bid_configuration
            ),
            main_price_scenario=self.main_price_scenario,
            price_scenarios=self.price_scenarios,
        )


class PriceAreaDayAhead(PriceArea):
    """This represents the reading version of price area day ahead.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area day ahead.
        data_record: The data record of the price area day ahead node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        default_bid_configuration: TODO
        main_price_scenario: TODO
        price_scenarios: TODO
    """

    node_type: Union[dm.DirectRelationReference, None] = None
    default_bid_configuration: Union[BidConfigurationDayAhead, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="defaultBidConfiguration"
    )
    main_price_scenario: Union[TimeSeries, str, None] = Field(None, alias="mainPriceScenario")
    price_scenarios: Union[list[TimeSeries], list[str], None] = Field(None, alias="priceScenarios")

    def as_write(self) -> PriceAreaDayAheadWrite:
        """Convert this read version of price area day ahead to the writing version."""
        return PriceAreaDayAheadWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
            asset_type=self.asset_type,
            default_bid_configuration=(
                self.default_bid_configuration.as_write()
                if isinstance(self.default_bid_configuration, DomainModel)
                else self.default_bid_configuration
            ),
            main_price_scenario=self.main_price_scenario,
            price_scenarios=self.price_scenarios,
        )

    def as_apply(self) -> PriceAreaDayAheadWrite:
        """Convert this read version of price area day ahead to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaDayAheadWrite(PriceAreaWrite):
    """This represents the writing version of price area day ahead.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the price area day ahead.
        data_record: The data record of the price area day ahead node.
        name: Name for the PowerAsset
        display_name: Display name for the PowerAsset.
        ordering: The ordering of the asset
        asset_type: The type of the asset
        default_bid_configuration: TODO
        main_price_scenario: TODO
        price_scenarios: TODO
    """

    node_type: Union[dm.DirectRelationReference, None] = None
    default_bid_configuration: Union[BidConfigurationDayAheadWrite, str, dm.NodeId, None] = Field(
        default=None, repr=False, alias="defaultBidConfiguration"
    )
    main_price_scenario: Union[TimeSeries, str, None] = Field(None, alias="mainPriceScenario")
    price_scenarios: Union[list[TimeSeries], list[str], None] = Field(None, alias="priceScenarios")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            PriceAreaDayAhead, dm.ViewId("power_ops_core", "PriceAreaDayAhead", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.display_name is not None or write_none:
            properties["displayName"] = self.display_name

        if self.ordering is not None or write_none:
            properties["ordering"] = self.ordering

        if self.asset_type is not None or write_none:
            properties["assetType"] = self.asset_type

        if self.default_bid_configuration is not None:
            properties["defaultBidConfiguration"] = {
                "space": (
                    self.space
                    if isinstance(self.default_bid_configuration, str)
                    else self.default_bid_configuration.space
                ),
                "externalId": (
                    self.default_bid_configuration
                    if isinstance(self.default_bid_configuration, str)
                    else self.default_bid_configuration.external_id
                ),
            }

        if self.main_price_scenario is not None or write_none:
            properties["mainPriceScenario"] = (
                self.main_price_scenario
                if isinstance(self.main_price_scenario, str) or self.main_price_scenario is None
                else self.main_price_scenario.external_id
            )

        if self.price_scenarios is not None or write_none:
            properties["priceScenarios"] = [
                value if isinstance(value, str) else value.external_id for value in self.price_scenarios or []
            ] or None

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.default_bid_configuration, DomainModelWrite):
            other_resources = self.default_bid_configuration._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.main_price_scenario, CogniteTimeSeries):
            resources.time_series.append(self.main_price_scenario)

        if isinstance(self.price_scenarios, CogniteTimeSeries):
            resources.time_series.append(self.price_scenarios)

        return resources


class PriceAreaDayAheadApply(PriceAreaDayAheadWrite):
    def __new__(cls, *args, **kwargs) -> PriceAreaDayAheadApply:
        warnings.warn(
            "PriceAreaDayAheadApply is deprecated and will be removed in v1.0. Use PriceAreaDayAheadWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "PriceAreaDayAhead.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class PriceAreaDayAheadList(DomainModelList[PriceAreaDayAhead]):
    """List of price area day aheads in the read version."""

    _INSTANCE = PriceAreaDayAhead

    def as_write(self) -> PriceAreaDayAheadWriteList:
        """Convert these read versions of price area day ahead to the writing versions."""
        return PriceAreaDayAheadWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> PriceAreaDayAheadWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class PriceAreaDayAheadWriteList(DomainModelWriteList[PriceAreaDayAheadWrite]):
    """List of price area day aheads in the writing version."""

    _INSTANCE = PriceAreaDayAheadWrite


class PriceAreaDayAheadApplyList(PriceAreaDayAheadWriteList): ...


def _create_price_area_day_ahead_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    min_ordering: int | None = None,
    max_ordering: int | None = None,
    asset_type: str | list[str] | None = None,
    asset_type_prefix: str | None = None,
    default_bid_configuration: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if min_ordering is not None or max_ordering is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("ordering"), gte=min_ordering, lte=max_ordering))
    if isinstance(asset_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("assetType"), value=asset_type))
    if asset_type and isinstance(asset_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("assetType"), values=asset_type))
    if asset_type_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("assetType"), value=asset_type_prefix))
    if default_bid_configuration and isinstance(default_bid_configuration, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("defaultBidConfiguration"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": default_bid_configuration},
            )
        )
    if default_bid_configuration and isinstance(default_bid_configuration, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("defaultBidConfiguration"),
                value={"space": default_bid_configuration[0], "externalId": default_bid_configuration[1]},
            )
        )
    if (
        default_bid_configuration
        and isinstance(default_bid_configuration, list)
        and isinstance(default_bid_configuration[0], str)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("defaultBidConfiguration"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in default_bid_configuration],
            )
        )
    if (
        default_bid_configuration
        and isinstance(default_bid_configuration, list)
        and isinstance(default_bid_configuration[0], tuple)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("defaultBidConfiguration"),
                values=[{"space": item[0], "externalId": item[1]} for item in default_bid_configuration],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from math import floor, log10
from typing import Any

import pandas as pd
from cognite.client.data_classes import Sequence

from cognite.powerops.resync.models.base import CDFSequence

p_min_fallback = 0.0
p_max_fallback = 100_000_000_000_000_000_00.0
head_loss_factor_fallback = 0.0
connection_losses_fallback = 0.0
inlet_reservoir_fallback = ""


def round_sig(x: float, sig: int = 2):
    return round(x, sig - int(floor(log10(abs(x)))) - 1)


def _create_generator_efficiency_curve(
    generator_attributes: dict[str, Any], generator_name: str, generator_external_id: str
) -> CDFSequence:
    x_col_name = "generator_power"
    y_col_name = "generator_efficiency"
    sequence = Sequence(
        external_id=f"{generator_external_id}_generator_efficiency_curve",
        name=f"{generator_name} generator efficiency curve",
        columns=[{"valueType": "DOUBLE", "externalId": x_col_name}, {"valueType": "DOUBLE", "externalId": y_col_name}],
    )
    data = generator_attributes["gen_eff_curve"]
    efficiency_curve = CDFSequence(
        sequence=sequence, content=pd.DataFrame({x_col_name: data["x"], y_col_name: data["y"]}, dtype=float)
    )
    return efficiency_curve


def _create_turbine_efficiency_curve(
    generator_attributes: dict[str, Any], generator_name: str, generator_external_id: str
) -> CDFSequence:
    data = generator_attributes["turb_eff_curves"]
    ref_col_name = "head"
    x_col_name = "flow"
    y_col_name = "turbine_efficiency"
    sequence = Sequence(
        external_id=f"{generator_external_id}_turbine_efficiency_curve",
        name=f"{generator_name} turbine efficiency curve",
        columns=[
            {"valueType": "DOUBLE", "externalId": ref_col_name},
            {"valueType": "DOUBLE", "externalId": x_col_name},
            {"valueType": "DOUBLE", "externalId": y_col_name},
        ],
    )
    efficiency_df = pd.DataFrame(
        {
            ref_col_name: [entry["ref"] for entry in data for _ in range(len(entry["x"]))],
            x_col_name: [item for entry in data for item in entry["x"]],
            y_col_name: [item for entry in data for item in entry["y"]],
        },
        dtype=float,
    )
    turbine_efficiency_curve = CDFSequence(sequence=sequence, content=efficiency_df)
    return turbine_efficiency_curve


def _get_single_value(value_or_time_series: float | dict) -> float:
    """Get the single value from a time series, or a value
    returns the value if value_or_time_series is a value, otherwise the first value in the time series

    Parameters
    ----------
    value_or_time_series : float | dict
        Either a simple numeric value, or a dictionary (time series, {datetime string: value})
    """
    if isinstance(value_or_time_series, dict):
        return next(iter(value_or_time_series.values()))
    return value_or_time_series


def _plant_to_inlet_reservoir_with_losses(
    plant_name: str, all_connections: list[dict], all_junctions: dict, all_tunnels: dict, reservoirs: set[str]
) -> tuple[str, float]:
    """Search for a reservoir connected to a plant, starting from the plant and searching breadth first.

    Parameters
    ----------
    plant_name : str
        The plant we want to find a connection from
    all_connections : list[dict]
        All connections in the model.
    reservoirs : dict
        All reservoirs in the model. Keys are reservoir names.

    Returns
    -------
    Optional[str]
        The name of the reservoir connected to the plant, or None if no reservoir was found.
    """

    def get_connection_path_from_last_visited(visited_paths: list, last_visited_id: int) -> list:
        """Return the correct sequence of visited connections based on the last visited connection among
        the list of connection paths visited

            Parameters
            ----------
            visited_paths : list
                 A list that holds the lists of each visited connection path. A connection path will be a list of
                 connection IDs, e.g. [1,2,3] means connection with ID 1 was first visited, then 2, then 3
            last_visited_id: int
                The ID of the connection that was last visited. This will be the last item in one of the lists of
                visited paths

            Returns
            -------
            list
                The path or sequence of connections that has the last_visited_id as its last visited connection among
                the visited_paths
        """
        for connection in visited_paths:
            if connection[-1] == last_visited_id:
                connection_path_index = visited_paths.index(connection)
                return visited_paths[connection_path_index]

    def calculate_losses_from_connection_path(
        all_junctions: dict, all_tunnels: dict, connection_by_id: int, connection_path: list[int]
    ):
        """Loop through connections in connection path, retrieve the losses for that connection among the all_juntions
        or all_tunnels based on the type of connection, and sum up the total losses from the connection path
        """
        sum_losses = 0
        order_to_loss_factor_key = {0: "loss_factor_1", 1: "loss_factor_2"}
        for connection_id in connection_path:
            connection = connection_by_id[connection_id]
            connection_name = connection["to"]
            if connection_name in all_junctions:
                if connection.get("order") in order_to_loss_factor_key:
                    junction_losses = all_junctions[connection_name]
                    loss_order = connection["order"]
                    sum_losses += junction_losses[order_to_loss_factor_key[loss_order]]
            elif connection_name in all_tunnels:
                tunnel_loss = all_tunnels[connection_name].get("loss_factor", 0)
                sum_losses += tunnel_loss
        return sum_losses

    queue = []
    connection_by_id = dict(enumerate(all_connections))
    track_connection_paths = []
    last_connection_id = None

    for connection_id, connection in connection_by_id.items():
        if (
            connection["to"] == plant_name and connection.get("to_type", "plant") == "plant"
        ):  # if to_type is specified, it must be "plant"
            queue.append((connection_id, connection))
            track_connection_paths.append([connection_id])  # add the first connection to the path
            break
    visited = []
    while queue:
        connection_id, connection = queue.pop(0)
        if connection not in visited:
            # Check if the given connection is from a reservoir
            # If we have "from_type" we can check directly if the object is a reservoir
            try:
                if connection["from_type"] == "reservoir":
                    inlet_reservoir = connection["from"]
                    last_connection_id = connection_id
                    break
            # If we don't have "from_type" we have to check if the name of the object is in the
            # list of reservoirs
            except KeyError:
                if connection["from"] in reservoirs:
                    inlet_reservoir = connection["from"]
                    last_connection_id = connection_id
                    break

            visited.append(connection)
            for candidate_connection_id, candidate_connection in connection_by_id.items():
                # if the candidate connection is extension from the current connection, traverse it
                if candidate_connection["to"] == connection["from"]:
                    queue.append((candidate_connection_id, candidate_connection))
                    new_path_list = get_connection_path_from_last_visited(track_connection_paths, connection_id)
                    track_connection_paths.append([*new_path_list, candidate_connection_id])

    if last_connection_id is None:
        return (inlet_reservoir_fallback, connection_losses_fallback)  # TODO: raise an error here instead?

    connection_path = get_connection_path_from_last_visited(track_connection_paths, last_connection_id)

    connection_losses = calculate_losses_from_connection_path(
        all_junctions, all_tunnels, connection_by_id, connection_path
    )

    return (inlet_reservoir, round_sig(connection_losses, 4) if connection_losses else connection_losses)



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "ProcessingLog",
    "ProcessingLogApply",
    "ProcessingLogList",
    "ProcessingLogApplyList",
    "ProcessingLogFields",
    "ProcessingLogTextFields",
]


ProcessingLogTextFields = Literal["state", "timestamp", "error_message"]
ProcessingLogFields = Literal["state", "timestamp", "error_message"]

_PROCESSINGLOG_PROPERTIES_BY_FIELD = {
    "state": "state",
    "timestamp": "timestamp",
    "error_message": "errorMessage",
}


class ProcessingLog(DomainModel):
    space: str = "cogShop"
    state: Optional[str] = None
    timestamp: Optional[str] = None
    error_message: Optional[str] = Field(None, alias="errorMessage")

    def as_apply(self) -> ProcessingLogApply:
        return ProcessingLogApply(
            external_id=self.external_id,
            state=self.state,
            timestamp=self.timestamp,
            error_message=self.error_message,
        )


class ProcessingLogApply(DomainModelApply):
    space: str = "cogShop"
    state: Optional[str] = None
    timestamp: Optional[str] = None
    error_message: Optional[str] = Field(None, alias="errorMessage")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.state is not None:
            properties["state"] = self.state
        if self.timestamp is not None:
            properties["timestamp"] = self.timestamp
        if self.error_message is not None:
            properties["errorMessage"] = self.error_message
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("cogShop", "ProcessingLog"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class ProcessingLogList(TypeList[ProcessingLog]):
    _NODE = ProcessingLog

    def as_apply(self) -> ProcessingLogApplyList:
        return ProcessingLogApplyList([node.as_apply() for node in self.data])


class ProcessingLogApplyList(TypeApplyList[ProcessingLogApply]):
    _NODE = ProcessingLogApply



from __future__ import annotations

from abc import ABC
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, Callable, ClassVar, Literal, Optional, TypeVar, Union

from cognite.client.data_classes import Asset, AssetList, LabelDefinition, LabelDefinitionList, Relationship, TimeSeries
from typing_extensions import Self

from cognite.powerops.cdf_labels import AssetLabel, RelationshipLabel
from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.resync.models.base.asset_type import AssetType
from cognite.powerops.resync.models.base.cdf_resources import CDFFile, CDFSequence
from cognite.powerops.resync.models.base.model import Model
from cognite.powerops.utils.serialization import get_pydantic_annotation


class AssetModel(Model, ABC, validate_assignment=True):
    root_asset: ClassVar[Optional[Asset]] = None

    def assets(self) -> list[Asset]:
        return [item.as_asset() for item in self._asset_types()]

    def relationships(self) -> list[Relationship]:
        return [edge for item in self._asset_types() for edge in item.relationships()]

    def parent_assets(self, include_root: bool = True) -> AssetList:
        if not self.root_asset:
            return AssetList([])

        def _to_name(external_id: str) -> str:
            parts = external_id.replace("_", " ").split(" ")
            if parts[0].lower() == "rkom":
                parts[0] = parts[0].upper()
            else:
                parts[0] = parts[0].title()
            # The replacing is to have avoid doing changes, even though this is inconsistent with the other parent
            # assets.
            return " ".join(parts).replace("Bid process", "Bid").replace("bid process", "bid")

        parent_and_description_ids = set()
        for field in self.model_fields.values():
            annotation, outer = get_pydantic_annotation(field.annotation, self)
            if issubclass(annotation, AssetType) and outer is list:
                parent_and_description_ids.add((annotation.parent_external_id, annotation.parent_description or ""))

        return AssetList(
            sorted(
                ([self.root_asset] if include_root else [])
                + [
                    Asset(
                        external_id=parent_id,
                        name=_to_name(parent_id),
                        parent_external_id=self.root_asset.external_id,
                        description=description,
                    )
                    for parent_id, description in parent_and_description_ids
                ],
                key=lambda asset: asset.external_id,
            )
        )

    def labels(self) -> list[LabelDefinition]:
        # for asset in assets:
        # for relationship in relationships:
        # return [
        #     LabelDefinition(external_id=external_id, name=external_id) for external_id in sorted(label_external_ids)
        # Labels are for the entire resync.
        return AssetLabel.as_label_definitions() + RelationshipLabel.as_label_definitions()

    cdf_resources: ClassVar[dict[Union[Callable, tuple[Callable, str]], type]] = {
        **dict(Model.cdf_resources.items()),
        assets: Asset,
        relationships: Relationship,
        parent_assets: Asset,
        labels: LabelDefinition,
    }

    # Need to set classmethod here to have access to the underlying function in cdf_resources
    parent_assets = classmethod(parent_assets)

    @classmethod
    def load_from_cdf_resources(
        cls: type[Self], data: dict[str, Any], link: Literal["external_id", "object"] = "object"
    ) -> Self:
        if not data.get("assets"):
            return cls()
        loaded_by_type_external_id = cls._load_by_type_external_id(data)

        parsed = cls._create_cls_arguments(loaded_by_type_external_id)

        instance = cls(**parsed)

        cls._set_linked_resources(loaded_by_type_external_id)
        return instance

    @classmethod
    def from_cdf(cls: type[T_Asset_Model], client: PowerOpsClient, data_set_external_id: str) -> T_Asset_Model:
        cdf = client.cdf
        parent_assets = cls.parent_assets(include_root=False)

        assets = cdf.assets.list(
            asset_subtree_external_ids=parent_assets.as_external_ids(),
            limit=-1,
            data_set_external_ids=[data_set_external_id],
        )
        if not assets:
            return cls()

        relationships = cdf.relationships.list(
            source_external_ids=assets.as_external_ids(),
            source_types=["asset"],
            target_types=["timeseries", "asset", "sequence", "file"],
            data_set_external_ids=[data_set_external_id],
            limit=-1,
        )
        # TimeSeries are a special case as resync are only referring to them by external id.
        # So we do not need to fetch them
        time_series = [
            TimeSeries(external_id=relationship.target_external_id)
            for relationship in relationships
            if relationship.target_type.casefold() == "timeseries"
        ]
        sequence_ids = [
            relationship.target_external_id
            for relationship in relationships
            if relationship.target_type.casefold() == "sequence"
        ]
        if sequence_ids:
            sequences = cdf.sequences.retrieve_multiple(external_ids=sequence_ids, ignore_unknown_ids=True)
            cdf_sequences = [CDFSequence(sequence=sequence) for sequence in sequences]
            cls._add_missing_hash(cdf, cdf_sequences)
        else:
            cdf_sequences = []

        file_ids = [
            relationship.target_external_id
            for relationship in relationships
            if relationship.target_type.casefold() == "file"
        ]
        if file_ids:
            files = cdf.files.retrieve_multiple(external_ids=file_ids, ignore_unknown_ids=True)
            cdf_files = [CDFFile(meta=file) for file in files]
            cls._add_missing_hash(cdf, cdf_files)
        else:
            cdf_files = []

        return cls.load_from_cdf_resources(
            {
                "assets": assets,
                "relationships": relationships,
                "sequences": cdf_sequences,
                "timeseries": time_series,
                "files": cdf_files,
            }
        )

    @classmethod
    def static_resources_from_cdf(cls, client: PowerOpsClient) -> dict[str, AssetList | LabelDefinitionList]:
        local_parent_assets = cls.parent_assets(include_root=True)
        parent_assets = client.cdf.assets.retrieve_multiple(
            external_ids=local_parent_assets.as_external_ids(), ignore_unknown_ids=True
        )
        labels = client.cdf.labels.list(data_set_external_ids=[client.datasets.read_dataset], limit=-1)
        return {"parent_assets": parent_assets, "labels": labels}

    def _asset_types(self) -> Iterable[AssetType]:
        yield from (item for item in self._resource_types() if isinstance(item, AssetType))

    @classmethod
    def _create_cls_arguments(cls, loaded_by_type_external_id: dict[str, dict[str, Any]]) -> dict[str, Any]:
        asset_by_parent_external_id = defaultdict(list)
        for asset in loaded_by_type_external_id["assets"].values():
            asset_by_parent_external_id[asset.parent_external_id].append(asset)
        arguments = {}
        asset_type_by_external_id = {}
        for field_name, field in cls.model_fields.items():
            annotation, outer = get_pydantic_annotation(field.annotation, cls)
            if issubclass(annotation, AssetType) and outer is list:
                assets = asset_by_parent_external_id.get(annotation.parent_external_id, [])
                # Hack to handle market case where two different asset types have the same parent
                if field_name == "nordpool_market":
                    arguments[field_name] = [
                        annotation.from_asset(asset)
                        for asset in assets
                        if asset.external_id == "market_configuration_nordpool_dayahead"
                    ]
                elif field_name == "rkom_market":
                    arguments[field_name] = [
                        annotation.from_asset(asset)
                        for asset in assets
                        if asset.external_id == "market_configuration_statnett_rkom_weekly"
                    ]
                else:
                    arguments[field_name] = [annotation.from_asset(asset) for asset in assets]
                asset_type_by_external_id.update({asset.external_id: asset for asset in arguments[field_name]})
            else:
                raise NotImplementedError()
        loaded_by_type_external_id["assets"] = asset_type_by_external_id
        return arguments

    @classmethod
    def _set_linked_resources(cls, loaded_by_type_external_id: dict[str, dict[str, Any]]) -> None:
        relationships_by_source_external_id = defaultdict(list)
        for relationship in loaded_by_type_external_id.get("relationships", {}).values():
            relationships_by_source_external_id[relationship.source_external_id].append(relationship)

        for source_id, relationships in relationships_by_source_external_id.items():
            if not (source := loaded_by_type_external_id["assets"].get(source_id)):
                # Todo print warning
                # Missing source
                continue

            for relationship in relationships:
                target_type = relationship.target_type.casefold()
                for key in (target_type + "s", target_type):
                    if target_items := loaded_by_type_external_id.get(key):
                        break
                else:
                    # Todo print warning
                    # Missing target type
                    continue
                if not (target := target_items.get(relationship.target_external_id)):
                    # Todo print warning
                    # Missing target external id
                    continue

                field_name = relationship.labels[0].external_id.split(".")[1]
                if field_name not in source.model_fields:
                    field_name += "s"
                if field_name not in source.model_fields:
                    field_name = field_name.rsplit("_", maxsplit=1)[0]
                if field_name not in source.model_fields and field_name == "reservoirs":
                    field_name = "inlet_reservoir"

                if field_name not in source.model_fields:
                    raise ValueError(f"Cannot find field {field_name} in {source}")

                annotation, outer = get_pydantic_annotation(source.model_fields[field_name].annotation, source)
                if outer is list:
                    getattr(source, field_name).append(target)
                elif outer is dict:
                    getattr(source, field_name)[target.external_id] = target
                elif outer is None:
                    setattr(source, field_name, target)
                else:
                    raise NotImplementedError()


T_Asset_Model = TypeVar("T_Asset_Model", bound=AssetModel)



from __future__ import annotations

from ._core import PriceScenario, PriceScenarioID, PriceScenarioV2
from .benchmarking import BenchmarkingConfig
from .dayahead import BidMatrixGeneratorConfig, BidProcessConfig
from .market import MARKET_BY_PRICE_AREA, MARKET_CONFIG_NORDPOOL_DAYAHEAD, Market
from .rkom import (
    Block,
    Product,
    ReserveScenario,
    ReserveScenarios,
    RKOMBidCombinationConfig,
    RKOMBidProcessConfig,
    RkomMarketConfig,
)

__all__ = [
    "PriceScenario",
    "PriceScenarioID",
    "PriceScenarioV2",
    "BenchmarkingConfig",
    "ReserveScenario",
    "ReserveScenarios",
    "RKOMBidCombinationConfig",
    "RKOMBidProcessConfig",
    "RkomMarketConfig",
    "Product",
    "Block",
    "BidMatrixGeneratorConfig",
    "BidProcessConfig",
    "Market",
    "MARKET_CONFIG_NORDPOOL_DAYAHEAD",
    "MARKET_BY_PRICE_AREA",
]



from __future__ import annotations

import json
from collections import Counter
from typing import ClassVar, Optional

from pydantic import BaseModel, Field, field_validator
from pydantic_core.core_schema import FieldValidationInfo

from cognite.powerops.resync.config.market import PriceScenarioID
from cognite.powerops.resync.config.market._core import Configuration, RelativeTime


class BidMatrixGeneratorConfig(BaseModel):
    name: str
    default_method: str
    default_function_external_id: str
    column_external_ids: ClassVar[list[str]] = [
        "shop_plant",
        "bid_matrix_generation_method",
        "function_external_id",
    ]


class BidProcessConfig(Configuration):
    name: str
    price_area_name: str = Field(alias="bid_price_area")
    price_scenarios: list[PriceScenarioID] = Field(alias="bid_price_scenarios")
    main_scenario: str = Field(alias="bid_main_scenario")
    bid_date: Optional[RelativeTime] = None
    shop_start: Optional[RelativeTime] = Field(None, alias="shop_starttime")
    shop_end: Optional[RelativeTime] = Field(None, alias="shop_endtime")
    bid_matrix_generator: str = Field(alias="bid_bid_matrix_generator_config_external_id")
    price_scenarios_per_watercourse: Optional[dict[str, set[str]]] = None
    is_default_config_for_price_area: bool = False
    no_shop: bool = Field(False, alias="no_shop")

    @field_validator("shop_start", "shop_end", "bid_date", mode="before")
    @classmethod
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value

    @field_validator("price_scenarios", mode="after")
    @classmethod
    def ensure_no_duplicates(cls, value: list[PriceScenarioID], info: FieldValidationInfo):
        scenario_ids = Counter(scenario.id for scenario in value)
        if duplicates := [scenario_id for scenario_id, count in scenario_ids.items() if count > 1]:
            bidprocess_name = info.data.get("name", "unknown")
            raise ValueError(
                f"Duplicate price scenarios for bidprocess {bidprocess_name} was found: {list(duplicates)}"
            )
        return value



# Mypy does not understand the pydantic classes that allows both alias and name to be used in population
# https://github.com/pydantic/pydantic/issues/3923
# mypy: disable-error-code="call-arg"
from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, Union

from cognite.client import CogniteClient
from cognite.client.data_classes import FileMetadataWrite
from cognite.client.exceptions import CogniteAPIError

from cognite.powerops.resync.v1.utils import ext_id_factory
from cognite.powerops.utils.serialization import load_yaml

logger = logging.getLogger(__name__)


@dataclass
class FileUploadConfiguration:
    """Data class for file upload configuration.

    Attributes:
        folder_path: The folder path to the files to upload.
        overwrite: Whether to overwrite existing files with the same name or external_id.
        file_metadata: A dictionary of file names to FileMetadataWrite objects.
    """

    folder_path: Path
    overwrite: bool
    file_metadata: dict[str, FileMetadataWrite]
    data_set_id: Optional[int] = None

    def __init__(
        self,
        data_set_id: Union[int, str],
        working_directory: Optional[Path] = None,
        folder_path: Optional[str] = None,
        overwrite: bool = True,
        file_metadata: Optional[list[dict[str, Any]]] = None,
        cdf_client: Optional[CogniteClient] = None,
    ) -> None:
        """Initializes the FileUploadConfiguration with the provided values and processes some of the values."""

        if folder_path:
            self.folder_path = Path(folder_path)
        elif working_directory:
            self.folder_path = working_directory / "files"
        else:
            raise ValueError("Either folder_path or working_directory must be provided")
        self.overwrite = overwrite

        if isinstance(data_set_id, str) and cdf_client:
            ds = cdf_client.data_sets.retrieve(external_id=data_set_id)

            self.data_set_id = ds.id if ds else None
        elif isinstance(data_set_id, int):
            self.data_set_id = data_set_id
        else:
            raise ValueError(f"Invalid data_set type {type(data_set_id)}")

        self.file_metadata = {}
        if file_metadata:
            for metadata in file_metadata:
                if "name" not in metadata:
                    raise ValueError("File metadata must contain a name field")
                if "directory" in metadata:
                    metadata["directory"] = self._convert_string_to_unix_path(metadata["directory"])

                file_metadata_instance = FileMetadataWrite(**metadata)
                if file_metadata_instance.name:
                    self.file_metadata[file_metadata_instance.name] = file_metadata_instance
                else:
                    raise ValueError("File metadata must contain a name field")

    @staticmethod
    def _convert_string_to_unix_path(path_str: str) -> str:
        path = Path(path_str).as_posix()

        return path if path.startswith("/") else f"/{path}"

    def upload_files_to_cdf(
        self,
        cdf_client: CogniteClient,
    ) -> list[str]:
        """Upload files to CDF.

        Given a file upload configuration, this function uploads the files to CDF. The function iterates over all files
        in the file configuration and uploads them to CDF. The function also checks if the file already exists in CDF
        and updates the file if it does. The function returns a list of all external ids for the uploaded files.

        Args:
            cdf_client: The Cognite client to use for file upload.
            file_configuration: The file upload configuration to use for file upload.
            directory_path: The directory path to the files to upload.
            data_set_id: The data set id to use for the files.

        Returns:
            file_external_ids: A list of all external ids for the uploaded files.

        Raises:
            CogniteAPIError: If the file already exists in CDF.
            ValueError: External id is required for file upload.
        """
        full_path = self.folder_path

        file_list = [file.name for file in full_path.iterdir() if file.is_file()]

        file_external_ids = []
        for file_name in file_list:
            file_metadata = self.file_metadata.get(
                file_name,
                FileMetadataWrite(
                    name=file_name,
                    external_id=ext_id_factory(FileMetadataWrite, {"name": file_name}),
                ),
            )

            file_metadata.data_set_id = file_metadata.data_set_id or self.data_set_id
            try:
                if file_metadata.external_id:
                    cdf_client.files.upload(
                        path=str(full_path / file_name),
                        overwrite=self.overwrite,
                        **file_metadata.dump(camel_case=False),
                    )
                    if self.overwrite:
                        cdf_client.files.update(file_metadata)
                    file_external_ids.append(file_metadata.external_id)
                else:
                    raise ValueError("External id is required for file upload")
            except CogniteAPIError as exc:
                if exc.code == 409:
                    logger.warning(f"File with external_id {file_metadata.external_id} already exists in CDF")
                else:
                    raise exc

        return file_external_ids


@dataclass
class PropertyConfiguration:
    """Data class for property configuration.

    Attributes:
        property: The property name.
        source_file: The source file to extract the property from.
        extraction_path: The extraction path in the source file.
        default_value: The default value if the property is not found.
        is_subtype: Whether the property is a subtype.
        is_list: Whether the property is a list.
        cast_type: The string representation of the cast type.
    """

    property: str
    source_file: Optional[Path]
    extraction_path: Optional[list[str]]
    default_value: Optional[Any]
    is_subtype: bool
    is_list: bool
    cast_type: Optional[str]

    def __init__(
        self,
        property: str,
        source_file: Optional[Path] = None,
        extraction_path: Optional[str] = None,
        default_value: Optional[Any] = None,
        is_subtype: bool = False,
        is_list: bool = False,
        cast_type: Optional[str] = None,
    ) -> None:
        """Initializes the PropertyConfiguration with the provided values and processes some of the values."""

        self.property = property
        self.source_file = source_file
        self.extraction_path = extraction_path.split(".") if extraction_path else None
        self.default_value = default_value
        self.is_subtype = is_subtype
        self.is_list = is_list
        self.cast_type = cast_type

    @classmethod
    def from_dictionary(cls, property_configurations: dict[str, Any]) -> list[PropertyConfiguration]:
        """Generate a list of PropertyConfiguration objects from a dictionary.

        Given a dictionary of property configurations, this function generates a list of PropertyConfiguration objects.

        Args:
            property_configurations: A dictionary of property configurations.

        Returns:
            A list of PropertyConfiguration objects.
        """

        parsed_property_configurations = []
        for property, configuration in property_configurations.items():
            parsed_property_configurations.append(cls(property=property, **configuration))
        return parsed_property_configurations

    @classmethod
    def from_string(
        cls, property: str, value: str, regex_pattern: str = r"^\[SOURCE:(.+.yaml)\](.+)?"
    ) -> PropertyConfiguration:
        """Generate a PropertyConfiguration object from a string matching a regex pattern.

        Given a property name and a string value, this function generates a PropertyConfiguration object if the string
        matches the regex pattern.

        Args:
            property: The property name.
            value: The string value to parse.
            regex_pattern: The regex pattern to match the string against.

        Returns:
            A PropertyConfiguration object.

        Raises:
            ValueError: If the string does not match the regex pattern.
        """

        match = re.match(regex_pattern, value)

        if match:
            source_file_path = Path(match.group(1))
            extraction_path = match.group(2)

            # TODO: implement default value
            default_value = None
            is_list = False
            cast_type = "str"
            is_subtype = False

            return cls(property, source_file_path, extraction_path, default_value, is_subtype, is_list, cast_type)
        else:
            raise ValueError(f"String {value} does not match pattern {regex_pattern}")

    @classmethod
    def from_data(cls, data: dict[str, Any]) -> list[PropertyConfiguration]:
        """Generate a list of PropertyConfiguration objects from a dictionary using a regex pattern.

        Given a dictionary of property configurations, this function generates a list of PropertyConfiguration objects
        for all properties that match the regex pattern.

        Args:
            data: A dictionary of property configurations.

        Returns:
            A list of PropertyConfiguration objects.
        """

        extraction_prefix = "[SOURCE"
        regex_pattern = r"\[SOURCE:(.+.yaml)\](.+)"

        fields_to_extract = []
        for property, value in data.items():
            if isinstance(value, str) and value.startswith(extraction_prefix):
                fields_to_extract.append(cls.from_string(property, value, regex_pattern))

        return fields_to_extract

    @staticmethod
    def union_lists(
        default_property_configurations: list[PropertyConfiguration],
        override_property_configurations: list[PropertyConfiguration],
    ) -> list[PropertyConfiguration]:
        """Combine two lists of PropertyConfiguration objects.

        Given two lists of PropertyConfiguration objects, this function combines the lists and removes duplicates. The
        default_property_configurations list is used as the base list, and the override_property_configurations list is
        used to override any configurations in the base list.

        Args:
            default_property_configurations: The base list of PropertyConfiguration objects.
            override_property_configurations: The list of PropertyConfiguration objects to override the base list.

        Returns:
            A list of PropertyConfiguration objects with no duplicates.
        """

        configurations_dict = {config.property: config for config in default_property_configurations}
        configurations_dict.update({config.property: config for config in override_property_configurations})

        return list(configurations_dict.values())


@dataclass
class DataModelConfiguration:
    """Data class for data model configuration.

    Attributes:
        name: The name of the data model type.
        type_: The type of the data model.
        properties: A dictionary of properties for the data model type.
        property_configurations: A list of PropertyConfiguration objects detailing the configuration of the properties.
    """

    name: str
    type_: type
    properties: dict[str, str]
    property_configurations: list[PropertyConfiguration]

    def __init__(
        self,
        name: str,
        type_: type,
        properties: Optional[dict[str, str]] = None,
        property_configurations: Optional[list[PropertyConfiguration]] = None,
    ) -> None:
        """Initializes the DataModelConfiguration with the provided values."""
        if property_configurations is None:
            property_configurations = []
        if property_configurations is None:
            property_configurations = []
        self.name = name
        self.type_ = type_

        if properties:
            self.properties = properties
        else:
            self.properties = self._get_properties_for_type(self.type_)

        self.property_configurations = property_configurations

    @classmethod
    def from_yaml(
        cls, all_write_classes: dict[str, type], configuration_path: Optional[Path] = None
    ) -> dict[str, DataModelConfiguration]:
        """Generate a dictionary of DataModelConfiguration objects from a YAML file.

        Given a path to a YAML file containing data model configurations, this function generates a dictionary of
        DataModelConfiguration objects.

        Args:
            configuration_path: The path to the YAML file containing data model configurations.

        Returns:
            A dictionary of DataModelConfiguration objects.

        Raises:
            ValueError: If a property configuration is not a valid property for the domain model type.
        """
        if configuration_path:
            raw_data_model_configuration = load_yaml(configuration_path, expected_return_type="dict")
        else:
            raw_data_model_configuration = {}

        all_data_model_configurations = {}
        for name, type_ in all_write_classes.items():
            data = raw_data_model_configuration.get(name, {})
            properties = cls._get_properties_for_type(type_)
            property_configurations = PropertyConfiguration.from_dictionary(data)

            for property_configuration in property_configurations:
                if property_configuration.property not in properties:
                    # TODO: make logging??
                    raise ValueError(
                        f"Property {property_configuration.property} not in domain model properties for {name}"
                    )

            all_data_model_configurations[name] = cls(
                type_=type_, name=name, properties=properties, property_configurations=property_configurations
            )

        return all_data_model_configurations

    @staticmethod
    def _get_properties_for_type(domain_model_type: type) -> dict[str, str]:
        """Get all properties for a domain model write class.

        Given a domain model write class, this function returns a dictionary with the property name as the key and the
        type as the value.

        Args:
            domain_model_type: The domain model write class to get the properties for.

        Returns:
            A dictionary with the property name as the key and the type as the value.
        """

        annotations = {"external_id": "str"}

        for domain_class in domain_model_type.__mro__:
            if domain_class.__name__.endswith("Write"):
                annotations.update(domain_class.__annotations__)

        return annotations



from __future__ import annotations

from collections.abc import Sequence
from typing import overload, Literal
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    Generator,
    GeneratorWrite,
    GeneratorFields,
    GeneratorList,
    GeneratorWriteList,
    GeneratorTextFields,
)
from cognite.powerops.client._generated.v1.data_classes._generator import (
    _GENERATOR_PROPERTIES_BY_FIELD,
    _create_generator_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .generator_turbine_efficiency_curves import GeneratorTurbineEfficiencyCurvesAPI
from .generator_start_stop_cost_time_series import GeneratorStartStopCostTimeSeriesAPI
from .generator_availability_time_series import GeneratorAvailabilityTimeSeriesAPI
from .generator_query import GeneratorQueryAPI


class GeneratorAPI(NodeAPI[Generator, GeneratorWrite, GeneratorList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[Generator]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Generator,
            class_list=GeneratorList,
            class_write_list=GeneratorWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.turbine_efficiency_curves_edge = GeneratorTurbineEfficiencyCurvesAPI(client)
        self.start_stop_cost_time_series = GeneratorStartStopCostTimeSeriesAPI(client, view_id)
        self.availability_time_series = GeneratorAvailabilityTimeSeriesAPI(client, view_id)

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> GeneratorQueryAPI[GeneratorList]:
        """Query starting at generators.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_penstock_number: The minimum value of the penstock number to filter on.
            max_penstock_number: The maximum value of the penstock number to filter on.
            min_start_stop_cost: The minimum value of the start stop cost to filter on.
            max_start_stop_cost: The maximum value of the start stop cost to filter on.
            generator_efficiency_curve: The generator efficiency curve to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generators to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for generators.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_generator_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_production_min,
            max_production_min,
            min_penstock_number,
            max_penstock_number,
            min_start_stop_cost,
            max_start_stop_cost,
            generator_efficiency_curve,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(GeneratorList)
        return GeneratorQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        generator: GeneratorWrite | Sequence[GeneratorWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) generators.

        Note: This method iterates through all nodes and timeseries linked to generator and creates them including the edges
        between the nodes. For example, if any of `turbine_efficiency_curves` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            generator: Generator or sequence of generators to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new generator:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import GeneratorWrite
                >>> client = PowerOpsModelsV1Client()
                >>> generator = GeneratorWrite(external_id="my_generator", ...)
                >>> result = client.generator.apply(generator)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.generator.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(generator, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more generator.

        Args:
            external_id: External id of the generator to delete.
            space: The space where all the generator are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete generator by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.generator.delete("my_generator")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.generator.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> Generator | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> GeneratorList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> Generator | GeneratorList | None:
        """Retrieve one or more generators by id(s).

        Args:
            external_id: External id or list of external ids of the generators.
            space: The space where all the generators are located.

        Returns:
            The requested generators.

        Examples:

            Retrieve generator by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> generator = client.generator.retrieve("my_generator")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.turbine_efficiency_curves_edge,
                    "turbine_efficiency_curves",
                    dm.DirectRelationReference("power_ops_types", "isSubAssetOf"),
                    "outwards",
                    dm.ViewId("power_ops_core", "TurbineEfficiencyCurve", "1"),
                ),
            ],
        )

    def search(
        self,
        query: str,
        properties: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> GeneratorList:
        """Search generators

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_penstock_number: The minimum value of the penstock number to filter on.
            max_penstock_number: The maximum value of the penstock number to filter on.
            min_start_stop_cost: The minimum value of the start stop cost to filter on.
            max_start_stop_cost: The maximum value of the start stop cost to filter on.
            generator_efficiency_curve: The generator efficiency curve to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generators to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results generators matching the query.

        Examples:

           Search for 'my_generator' in all text properties:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> generators = client.generator.search('my_generator')

        """
        filter_ = _create_generator_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_production_min,
            max_production_min,
            min_penstock_number,
            max_penstock_number,
            min_start_stop_cost,
            max_start_stop_cost,
            generator_efficiency_curve,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _GENERATOR_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: GeneratorFields | Sequence[GeneratorFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: GeneratorFields | Sequence[GeneratorFields] | None = None,
        group_by: GeneratorFields | Sequence[GeneratorFields] = None,
        query: str | None = None,
        search_properties: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: GeneratorFields | Sequence[GeneratorFields] | None = None,
        group_by: GeneratorFields | Sequence[GeneratorFields] | None = None,
        query: str | None = None,
        search_property: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across generators

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_penstock_number: The minimum value of the penstock number to filter on.
            max_penstock_number: The maximum value of the penstock number to filter on.
            min_start_stop_cost: The minimum value of the start stop cost to filter on.
            max_start_stop_cost: The maximum value of the start stop cost to filter on.
            generator_efficiency_curve: The generator efficiency curve to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generators to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count generators in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.generator.aggregate("count", space="my_space")

        """

        filter_ = _create_generator_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_production_min,
            max_production_min,
            min_penstock_number,
            max_penstock_number,
            min_start_stop_cost,
            max_start_stop_cost,
            generator_efficiency_curve,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _GENERATOR_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: GeneratorFields,
        interval: float,
        query: str | None = None,
        search_property: GeneratorTextFields | Sequence[GeneratorTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for generators

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_penstock_number: The minimum value of the penstock number to filter on.
            max_penstock_number: The maximum value of the penstock number to filter on.
            min_start_stop_cost: The minimum value of the start stop cost to filter on.
            max_start_stop_cost: The maximum value of the start stop cost to filter on.
            generator_efficiency_curve: The generator efficiency curve to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generators to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_generator_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_production_min,
            max_production_min,
            min_penstock_number,
            max_penstock_number,
            min_start_stop_cost,
            max_start_stop_cost,
            generator_efficiency_curve,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _GENERATOR_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        asset_type: str | list[str] | None = None,
        asset_type_prefix: str | None = None,
        min_production_min: float | None = None,
        max_production_min: float | None = None,
        min_penstock_number: int | None = None,
        max_penstock_number: int | None = None,
        min_start_stop_cost: float | None = None,
        max_start_stop_cost: float | None = None,
        generator_efficiency_curve: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        sort_by: GeneratorFields | Sequence[GeneratorFields] | None = None,
        direction: Literal["ascending", "descending"] = "ascending",
        retrieve_edges: bool = True,
    ) -> GeneratorList:
        """List/filter generators

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            asset_type: The asset type to filter on.
            asset_type_prefix: The prefix of the asset type to filter on.
            min_production_min: The minimum value of the production min to filter on.
            max_production_min: The maximum value of the production min to filter on.
            min_penstock_number: The minimum value of the penstock number to filter on.
            max_penstock_number: The maximum value of the penstock number to filter on.
            min_start_stop_cost: The minimum value of the start stop cost to filter on.
            max_start_stop_cost: The maximum value of the start stop cost to filter on.
            generator_efficiency_curve: The generator efficiency curve to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of generators to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            sort_by: The property to sort by.
            direction: The direction to sort by, either 'ascending' or 'descending'.
            retrieve_edges: Whether to retrieve `turbine_efficiency_curves` external ids for the generators. Defaults to True.

        Returns:
            List of requested generators

        Examples:

            List generators and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> generators = client.generator.list(limit=5)

        """
        filter_ = _create_generator_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            asset_type,
            asset_type_prefix,
            min_production_min,
            max_production_min,
            min_penstock_number,
            max_penstock_number,
            min_start_stop_cost,
            max_start_stop_cost,
            generator_efficiency_curve,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            properties_by_field=_GENERATOR_PROPERTIES_BY_FIELD,
            sort_by=sort_by,
            direction=direction,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.turbine_efficiency_curves_edge,
                    "turbine_efficiency_curves",
                    dm.DirectRelationReference("power_ops_types", "isSubAssetOf"),
                    "outwards",
                    dm.ViewId("power_ops_core", "TurbineEfficiencyCurve", "1"),
                ),
            ],
        )



from ._core import (
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    GraphQLList,
    ResourcesWrite,
    ResourcesWriteResult,
    PageInfo,
)
from ._alert import (
    Alert,
    AlertApply,
    AlertApplyList,
    AlertFields,
    AlertGraphQL,
    AlertList,
    AlertTextFields,
    AlertWrite,
    AlertWriteList,
)
from ._benchmarking_calculation_input import (
    BenchmarkingCalculationInput,
    BenchmarkingCalculationInputApply,
    BenchmarkingCalculationInputApplyList,
    BenchmarkingCalculationInputFields,
    BenchmarkingCalculationInputGraphQL,
    BenchmarkingCalculationInputList,
    BenchmarkingCalculationInputTextFields,
    BenchmarkingCalculationInputWrite,
    BenchmarkingCalculationInputWriteList,
)
from ._benchmarking_calculation_output import (
    BenchmarkingCalculationOutput,
    BenchmarkingCalculationOutputApply,
    BenchmarkingCalculationOutputApplyList,
    BenchmarkingCalculationOutputFields,
    BenchmarkingCalculationOutputGraphQL,
    BenchmarkingCalculationOutputList,
    BenchmarkingCalculationOutputTextFields,
    BenchmarkingCalculationOutputWrite,
    BenchmarkingCalculationOutputWriteList,
)
from ._benchmarking_configuration_day_ahead import (
    BenchmarkingConfigurationDayAhead,
    BenchmarkingConfigurationDayAheadApply,
    BenchmarkingConfigurationDayAheadApplyList,
    BenchmarkingConfigurationDayAheadFields,
    BenchmarkingConfigurationDayAheadGraphQL,
    BenchmarkingConfigurationDayAheadList,
    BenchmarkingConfigurationDayAheadTextFields,
    BenchmarkingConfigurationDayAheadWrite,
    BenchmarkingConfigurationDayAheadWriteList,
)
from ._benchmarking_production_obligation_day_ahead import (
    BenchmarkingProductionObligationDayAhead,
    BenchmarkingProductionObligationDayAheadApply,
    BenchmarkingProductionObligationDayAheadApplyList,
    BenchmarkingProductionObligationDayAheadFields,
    BenchmarkingProductionObligationDayAheadGraphQL,
    BenchmarkingProductionObligationDayAheadList,
    BenchmarkingProductionObligationDayAheadTextFields,
    BenchmarkingProductionObligationDayAheadWrite,
    BenchmarkingProductionObligationDayAheadWriteList,
)
from ._benchmarking_result_day_ahead import (
    BenchmarkingResultDayAhead,
    BenchmarkingResultDayAheadApply,
    BenchmarkingResultDayAheadApplyList,
    BenchmarkingResultDayAheadFields,
    BenchmarkingResultDayAheadGraphQL,
    BenchmarkingResultDayAheadList,
    BenchmarkingResultDayAheadTextFields,
    BenchmarkingResultDayAheadWrite,
    BenchmarkingResultDayAheadWriteList,
)
from ._benchmarking_shop_case import (
    BenchmarkingShopCase,
    BenchmarkingShopCaseApply,
    BenchmarkingShopCaseApplyList,
    BenchmarkingShopCaseFields,
    BenchmarkingShopCaseGraphQL,
    BenchmarkingShopCaseList,
    BenchmarkingShopCaseWrite,
    BenchmarkingShopCaseWriteList,
)
from ._benchmarking_task_dispatcher_input_day_ahead import (
    BenchmarkingTaskDispatcherInputDayAhead,
    BenchmarkingTaskDispatcherInputDayAheadApply,
    BenchmarkingTaskDispatcherInputDayAheadApplyList,
    BenchmarkingTaskDispatcherInputDayAheadFields,
    BenchmarkingTaskDispatcherInputDayAheadGraphQL,
    BenchmarkingTaskDispatcherInputDayAheadList,
    BenchmarkingTaskDispatcherInputDayAheadTextFields,
    BenchmarkingTaskDispatcherInputDayAheadWrite,
    BenchmarkingTaskDispatcherInputDayAheadWriteList,
)
from ._benchmarking_task_dispatcher_output_day_ahead import (
    BenchmarkingTaskDispatcherOutputDayAhead,
    BenchmarkingTaskDispatcherOutputDayAheadApply,
    BenchmarkingTaskDispatcherOutputDayAheadApplyList,
    BenchmarkingTaskDispatcherOutputDayAheadFields,
    BenchmarkingTaskDispatcherOutputDayAheadGraphQL,
    BenchmarkingTaskDispatcherOutputDayAheadList,
    BenchmarkingTaskDispatcherOutputDayAheadTextFields,
    BenchmarkingTaskDispatcherOutputDayAheadWrite,
    BenchmarkingTaskDispatcherOutputDayAheadWriteList,
)
from ._bid_configuration_day_ahead import (
    BidConfigurationDayAhead,
    BidConfigurationDayAheadApply,
    BidConfigurationDayAheadApplyList,
    BidConfigurationDayAheadFields,
    BidConfigurationDayAheadGraphQL,
    BidConfigurationDayAheadList,
    BidConfigurationDayAheadTextFields,
    BidConfigurationDayAheadWrite,
    BidConfigurationDayAheadWriteList,
)
from ._bid_document import (
    BidDocument,
    BidDocumentApply,
    BidDocumentApplyList,
    BidDocumentFields,
    BidDocumentGraphQL,
    BidDocumentList,
    BidDocumentTextFields,
    BidDocumentWrite,
    BidDocumentWriteList,
)
from ._bid_document_afrr import (
    BidDocumentAFRR,
    BidDocumentAFRRApply,
    BidDocumentAFRRApplyList,
    BidDocumentAFRRFields,
    BidDocumentAFRRGraphQL,
    BidDocumentAFRRList,
    BidDocumentAFRRTextFields,
    BidDocumentAFRRWrite,
    BidDocumentAFRRWriteList,
)
from ._bid_document_day_ahead import (
    BidDocumentDayAhead,
    BidDocumentDayAheadApply,
    BidDocumentDayAheadApplyList,
    BidDocumentDayAheadFields,
    BidDocumentDayAheadGraphQL,
    BidDocumentDayAheadList,
    BidDocumentDayAheadTextFields,
    BidDocumentDayAheadWrite,
    BidDocumentDayAheadWriteList,
)
from ._bid_matrix import (
    BidMatrix,
    BidMatrixApply,
    BidMatrixApplyList,
    BidMatrixFields,
    BidMatrixGraphQL,
    BidMatrixList,
    BidMatrixTextFields,
    BidMatrixWrite,
    BidMatrixWriteList,
)
from ._bid_matrix_information import (
    BidMatrixInformation,
    BidMatrixInformationApply,
    BidMatrixInformationApplyList,
    BidMatrixInformationFields,
    BidMatrixInformationGraphQL,
    BidMatrixInformationList,
    BidMatrixInformationTextFields,
    BidMatrixInformationWrite,
    BidMatrixInformationWriteList,
)
from ._bid_row import (
    BidRow,
    BidRowApply,
    BidRowApplyList,
    BidRowFields,
    BidRowGraphQL,
    BidRowList,
    BidRowTextFields,
    BidRowWrite,
    BidRowWriteList,
)
from ._date_specification import (
    DateSpecification,
    DateSpecificationApply,
    DateSpecificationApplyList,
    DateSpecificationFields,
    DateSpecificationGraphQL,
    DateSpecificationList,
    DateSpecificationTextFields,
    DateSpecificationWrite,
    DateSpecificationWriteList,
)
from ._function_input import (
    FunctionInput,
    FunctionInputApply,
    FunctionInputApplyList,
    FunctionInputFields,
    FunctionInputGraphQL,
    FunctionInputList,
    FunctionInputTextFields,
    FunctionInputWrite,
    FunctionInputWriteList,
)
from ._function_output import (
    FunctionOutput,
    FunctionOutputApply,
    FunctionOutputApplyList,
    FunctionOutputFields,
    FunctionOutputGraphQL,
    FunctionOutputList,
    FunctionOutputTextFields,
    FunctionOutputWrite,
    FunctionOutputWriteList,
)
from ._generator import (
    Generator,
    GeneratorApply,
    GeneratorApplyList,
    GeneratorFields,
    GeneratorGraphQL,
    GeneratorList,
    GeneratorTextFields,
    GeneratorWrite,
    GeneratorWriteList,
)
from ._generator_efficiency_curve import (
    GeneratorEfficiencyCurve,
    GeneratorEfficiencyCurveApply,
    GeneratorEfficiencyCurveApplyList,
    GeneratorEfficiencyCurveFields,
    GeneratorEfficiencyCurveGraphQL,
    GeneratorEfficiencyCurveList,
    GeneratorEfficiencyCurveWrite,
    GeneratorEfficiencyCurveWriteList,
)
from ._market_configuration import (
    MarketConfiguration,
    MarketConfigurationApply,
    MarketConfigurationApplyList,
    MarketConfigurationFields,
    MarketConfigurationGraphQL,
    MarketConfigurationList,
    MarketConfigurationTextFields,
    MarketConfigurationWrite,
    MarketConfigurationWriteList,
)
from ._multi_scenario_partial_bid_matrix_calculation_input import (
    MultiScenarioPartialBidMatrixCalculationInput,
    MultiScenarioPartialBidMatrixCalculationInputApply,
    MultiScenarioPartialBidMatrixCalculationInputApplyList,
    MultiScenarioPartialBidMatrixCalculationInputFields,
    MultiScenarioPartialBidMatrixCalculationInputGraphQL,
    MultiScenarioPartialBidMatrixCalculationInputList,
    MultiScenarioPartialBidMatrixCalculationInputTextFields,
    MultiScenarioPartialBidMatrixCalculationInputWrite,
    MultiScenarioPartialBidMatrixCalculationInputWriteList,
)
from ._partial_bid_configuration import (
    PartialBidConfiguration,
    PartialBidConfigurationApply,
    PartialBidConfigurationApplyList,
    PartialBidConfigurationFields,
    PartialBidConfigurationGraphQL,
    PartialBidConfigurationList,
    PartialBidConfigurationTextFields,
    PartialBidConfigurationWrite,
    PartialBidConfigurationWriteList,
)
from ._partial_bid_matrix_calculation_input import (
    PartialBidMatrixCalculationInput,
    PartialBidMatrixCalculationInputApply,
    PartialBidMatrixCalculationInputApplyList,
    PartialBidMatrixCalculationInputFields,
    PartialBidMatrixCalculationInputGraphQL,
    PartialBidMatrixCalculationInputList,
    PartialBidMatrixCalculationInputTextFields,
    PartialBidMatrixCalculationInputWrite,
    PartialBidMatrixCalculationInputWriteList,
)
from ._partial_bid_matrix_calculation_output import (
    PartialBidMatrixCalculationOutput,
    PartialBidMatrixCalculationOutputApply,
    PartialBidMatrixCalculationOutputApplyList,
    PartialBidMatrixCalculationOutputFields,
    PartialBidMatrixCalculationOutputGraphQL,
    PartialBidMatrixCalculationOutputList,
    PartialBidMatrixCalculationOutputTextFields,
    PartialBidMatrixCalculationOutputWrite,
    PartialBidMatrixCalculationOutputWriteList,
)
from ._partial_bid_matrix_information import (
    PartialBidMatrixInformation,
    PartialBidMatrixInformationApply,
    PartialBidMatrixInformationApplyList,
    PartialBidMatrixInformationFields,
    PartialBidMatrixInformationGraphQL,
    PartialBidMatrixInformationList,
    PartialBidMatrixInformationTextFields,
    PartialBidMatrixInformationWrite,
    PartialBidMatrixInformationWriteList,
)
from ._partial_bid_matrix_information_with_scenarios import (
    PartialBidMatrixInformationWithScenarios,
    PartialBidMatrixInformationWithScenariosApply,
    PartialBidMatrixInformationWithScenariosApplyList,
    PartialBidMatrixInformationWithScenariosFields,
    PartialBidMatrixInformationWithScenariosGraphQL,
    PartialBidMatrixInformationWithScenariosList,
    PartialBidMatrixInformationWithScenariosTextFields,
    PartialBidMatrixInformationWithScenariosWrite,
    PartialBidMatrixInformationWithScenariosWriteList,
)
from ._plant import (
    Plant,
    PlantApply,
    PlantApplyList,
    PlantFields,
    PlantGraphQL,
    PlantList,
    PlantTextFields,
    PlantWrite,
    PlantWriteList,
)
from ._plant_information import (
    PlantInformation,
    PlantInformationApply,
    PlantInformationApplyList,
    PlantInformationFields,
    PlantInformationGraphQL,
    PlantInformationList,
    PlantInformationTextFields,
    PlantInformationWrite,
    PlantInformationWriteList,
)
from ._plant_water_value_based import (
    PlantWaterValueBased,
    PlantWaterValueBasedApply,
    PlantWaterValueBasedApplyList,
    PlantWaterValueBasedFields,
    PlantWaterValueBasedGraphQL,
    PlantWaterValueBasedList,
    PlantWaterValueBasedTextFields,
    PlantWaterValueBasedWrite,
    PlantWaterValueBasedWriteList,
)
from ._power_asset import (
    PowerAsset,
    PowerAssetApply,
    PowerAssetApplyList,
    PowerAssetFields,
    PowerAssetGraphQL,
    PowerAssetList,
    PowerAssetTextFields,
    PowerAssetWrite,
    PowerAssetWriteList,
)
from ._price_area import (
    PriceArea,
    PriceAreaApply,
    PriceAreaApplyList,
    PriceAreaFields,
    PriceAreaGraphQL,
    PriceAreaList,
    PriceAreaTextFields,
    PriceAreaWrite,
    PriceAreaWriteList,
)
from ._price_area_afrr import (
    PriceAreaAFRR,
    PriceAreaAFRRApply,
    PriceAreaAFRRApplyList,
    PriceAreaAFRRFields,
    PriceAreaAFRRGraphQL,
    PriceAreaAFRRList,
    PriceAreaAFRRTextFields,
    PriceAreaAFRRWrite,
    PriceAreaAFRRWriteList,
)
from ._price_area_day_ahead import (
    PriceAreaDayAhead,
    PriceAreaDayAheadApply,
    PriceAreaDayAheadApplyList,
    PriceAreaDayAheadFields,
    PriceAreaDayAheadGraphQL,
    PriceAreaDayAheadList,
    PriceAreaDayAheadTextFields,
    PriceAreaDayAheadWrite,
    PriceAreaDayAheadWriteList,
)
from ._price_area_information import (
    PriceAreaInformation,
    PriceAreaInformationApply,
    PriceAreaInformationApplyList,
    PriceAreaInformationFields,
    PriceAreaInformationGraphQL,
    PriceAreaInformationList,
    PriceAreaInformationTextFields,
    PriceAreaInformationWrite,
    PriceAreaInformationWriteList,
)
from ._price_production import (
    PriceProduction,
    PriceProductionApply,
    PriceProductionApplyList,
    PriceProductionFields,
    PriceProductionGraphQL,
    PriceProductionList,
    PriceProductionTextFields,
    PriceProductionWrite,
    PriceProductionWriteList,
)
from ._shop_attribute_mapping import (
    ShopAttributeMapping,
    ShopAttributeMappingApply,
    ShopAttributeMappingApplyList,
    ShopAttributeMappingFields,
    ShopAttributeMappingGraphQL,
    ShopAttributeMappingList,
    ShopAttributeMappingTextFields,
    ShopAttributeMappingWrite,
    ShopAttributeMappingWriteList,
)
from ._shop_based_partial_bid_configuration import (
    ShopBasedPartialBidConfiguration,
    ShopBasedPartialBidConfigurationApply,
    ShopBasedPartialBidConfigurationApplyList,
    ShopBasedPartialBidConfigurationFields,
    ShopBasedPartialBidConfigurationGraphQL,
    ShopBasedPartialBidConfigurationList,
    ShopBasedPartialBidConfigurationTextFields,
    ShopBasedPartialBidConfigurationWrite,
    ShopBasedPartialBidConfigurationWriteList,
)
from ._shop_case import (
    ShopCase,
    ShopCaseApply,
    ShopCaseApplyList,
    ShopCaseFields,
    ShopCaseGraphQL,
    ShopCaseList,
    ShopCaseWrite,
    ShopCaseWriteList,
)
from ._shop_commands import (
    ShopCommands,
    ShopCommandsApply,
    ShopCommandsApplyList,
    ShopCommandsFields,
    ShopCommandsGraphQL,
    ShopCommandsList,
    ShopCommandsTextFields,
    ShopCommandsWrite,
    ShopCommandsWriteList,
)
from ._shop_file import (
    ShopFile,
    ShopFileApply,
    ShopFileApplyList,
    ShopFileFields,
    ShopFileGraphQL,
    ShopFileList,
    ShopFileTextFields,
    ShopFileWrite,
    ShopFileWriteList,
)
from ._shop_model import (
    ShopModel,
    ShopModelApply,
    ShopModelApplyList,
    ShopModelFields,
    ShopModelGraphQL,
    ShopModelList,
    ShopModelTextFields,
    ShopModelWrite,
    ShopModelWriteList,
)
from ._shop_model_with_assets import (
    ShopModelWithAssets,
    ShopModelWithAssetsApply,
    ShopModelWithAssetsApplyList,
    ShopModelWithAssetsGraphQL,
    ShopModelWithAssetsList,
    ShopModelWithAssetsWrite,
    ShopModelWithAssetsWriteList,
)
from ._shop_output_time_series_definition import (
    ShopOutputTimeSeriesDefinition,
    ShopOutputTimeSeriesDefinitionApply,
    ShopOutputTimeSeriesDefinitionApplyList,
    ShopOutputTimeSeriesDefinitionFields,
    ShopOutputTimeSeriesDefinitionGraphQL,
    ShopOutputTimeSeriesDefinitionList,
    ShopOutputTimeSeriesDefinitionTextFields,
    ShopOutputTimeSeriesDefinitionWrite,
    ShopOutputTimeSeriesDefinitionWriteList,
)
from ._shop_penalty_report import (
    ShopPenaltyReport,
    ShopPenaltyReportApply,
    ShopPenaltyReportApplyList,
    ShopPenaltyReportFields,
    ShopPenaltyReportGraphQL,
    ShopPenaltyReportList,
    ShopPenaltyReportTextFields,
    ShopPenaltyReportWrite,
    ShopPenaltyReportWriteList,
)
from ._shop_preprocessor_input import (
    ShopPreprocessorInput,
    ShopPreprocessorInputApply,
    ShopPreprocessorInputApplyList,
    ShopPreprocessorInputFields,
    ShopPreprocessorInputGraphQL,
    ShopPreprocessorInputList,
    ShopPreprocessorInputTextFields,
    ShopPreprocessorInputWrite,
    ShopPreprocessorInputWriteList,
)
from ._shop_preprocessor_output import (
    ShopPreprocessorOutput,
    ShopPreprocessorOutputApply,
    ShopPreprocessorOutputApplyList,
    ShopPreprocessorOutputFields,
    ShopPreprocessorOutputGraphQL,
    ShopPreprocessorOutputList,
    ShopPreprocessorOutputTextFields,
    ShopPreprocessorOutputWrite,
    ShopPreprocessorOutputWriteList,
)
from ._shop_result import (
    ShopResult,
    ShopResultApply,
    ShopResultApplyList,
    ShopResultFields,
    ShopResultGraphQL,
    ShopResultList,
    ShopResultWrite,
    ShopResultWriteList,
)
from ._shop_scenario import (
    ShopScenario,
    ShopScenarioApply,
    ShopScenarioApplyList,
    ShopScenarioFields,
    ShopScenarioGraphQL,
    ShopScenarioList,
    ShopScenarioTextFields,
    ShopScenarioWrite,
    ShopScenarioWriteList,
)
from ._shop_scenario_set import (
    ShopScenarioSet,
    ShopScenarioSetApply,
    ShopScenarioSetApplyList,
    ShopScenarioSetFields,
    ShopScenarioSetGraphQL,
    ShopScenarioSetList,
    ShopScenarioSetTextFields,
    ShopScenarioSetWrite,
    ShopScenarioSetWriteList,
)
from ._shop_time_series import (
    ShopTimeSeries,
    ShopTimeSeriesApply,
    ShopTimeSeriesApplyList,
    ShopTimeSeriesFields,
    ShopTimeSeriesGraphQL,
    ShopTimeSeriesList,
    ShopTimeSeriesTextFields,
    ShopTimeSeriesWrite,
    ShopTimeSeriesWriteList,
)
from ._shop_trigger_input import (
    ShopTriggerInput,
    ShopTriggerInputApply,
    ShopTriggerInputApplyList,
    ShopTriggerInputFields,
    ShopTriggerInputGraphQL,
    ShopTriggerInputList,
    ShopTriggerInputTextFields,
    ShopTriggerInputWrite,
    ShopTriggerInputWriteList,
)
from ._shop_trigger_output import (
    ShopTriggerOutput,
    ShopTriggerOutputApply,
    ShopTriggerOutputApplyList,
    ShopTriggerOutputFields,
    ShopTriggerOutputGraphQL,
    ShopTriggerOutputList,
    ShopTriggerOutputTextFields,
    ShopTriggerOutputWrite,
    ShopTriggerOutputWriteList,
)
from ._task_dispatcher_input import (
    TaskDispatcherInput,
    TaskDispatcherInputApply,
    TaskDispatcherInputApplyList,
    TaskDispatcherInputFields,
    TaskDispatcherInputGraphQL,
    TaskDispatcherInputList,
    TaskDispatcherInputTextFields,
    TaskDispatcherInputWrite,
    TaskDispatcherInputWriteList,
)
from ._task_dispatcher_output import (
    TaskDispatcherOutput,
    TaskDispatcherOutputApply,
    TaskDispatcherOutputApplyList,
    TaskDispatcherOutputFields,
    TaskDispatcherOutputGraphQL,
    TaskDispatcherOutputList,
    TaskDispatcherOutputTextFields,
    TaskDispatcherOutputWrite,
    TaskDispatcherOutputWriteList,
)
from ._total_bid_matrix_calculation_input import (
    TotalBidMatrixCalculationInput,
    TotalBidMatrixCalculationInputApply,
    TotalBidMatrixCalculationInputApplyList,
    TotalBidMatrixCalculationInputFields,
    TotalBidMatrixCalculationInputGraphQL,
    TotalBidMatrixCalculationInputList,
    TotalBidMatrixCalculationInputTextFields,
    TotalBidMatrixCalculationInputWrite,
    TotalBidMatrixCalculationInputWriteList,
)
from ._total_bid_matrix_calculation_output import (
    TotalBidMatrixCalculationOutput,
    TotalBidMatrixCalculationOutputApply,
    TotalBidMatrixCalculationOutputApplyList,
    TotalBidMatrixCalculationOutputFields,
    TotalBidMatrixCalculationOutputGraphQL,
    TotalBidMatrixCalculationOutputList,
    TotalBidMatrixCalculationOutputTextFields,
    TotalBidMatrixCalculationOutputWrite,
    TotalBidMatrixCalculationOutputWriteList,
)
from ._turbine_efficiency_curve import (
    TurbineEfficiencyCurve,
    TurbineEfficiencyCurveApply,
    TurbineEfficiencyCurveApplyList,
    TurbineEfficiencyCurveFields,
    TurbineEfficiencyCurveGraphQL,
    TurbineEfficiencyCurveList,
    TurbineEfficiencyCurveWrite,
    TurbineEfficiencyCurveWriteList,
)
from ._water_value_based_partial_bid_configuration import (
    WaterValueBasedPartialBidConfiguration,
    WaterValueBasedPartialBidConfigurationApply,
    WaterValueBasedPartialBidConfigurationApplyList,
    WaterValueBasedPartialBidConfigurationFields,
    WaterValueBasedPartialBidConfigurationGraphQL,
    WaterValueBasedPartialBidConfigurationList,
    WaterValueBasedPartialBidConfigurationTextFields,
    WaterValueBasedPartialBidConfigurationWrite,
    WaterValueBasedPartialBidConfigurationWriteList,
)
from ._water_value_based_partial_bid_matrix_calculation_input import (
    WaterValueBasedPartialBidMatrixCalculationInput,
    WaterValueBasedPartialBidMatrixCalculationInputApply,
    WaterValueBasedPartialBidMatrixCalculationInputApplyList,
    WaterValueBasedPartialBidMatrixCalculationInputFields,
    WaterValueBasedPartialBidMatrixCalculationInputGraphQL,
    WaterValueBasedPartialBidMatrixCalculationInputList,
    WaterValueBasedPartialBidMatrixCalculationInputTextFields,
    WaterValueBasedPartialBidMatrixCalculationInputWrite,
    WaterValueBasedPartialBidMatrixCalculationInputWriteList,
)
from ._watercourse import (
    Watercourse,
    WatercourseApply,
    WatercourseApplyList,
    WatercourseFields,
    WatercourseGraphQL,
    WatercourseList,
    WatercourseTextFields,
    WatercourseWrite,
    WatercourseWriteList,
)

BenchmarkingCalculationInput.model_rebuild()
BenchmarkingCalculationInputGraphQL.model_rebuild()
BenchmarkingCalculationInputWrite.model_rebuild()
BenchmarkingCalculationInputApply.model_rebuild()
BenchmarkingCalculationOutput.model_rebuild()
BenchmarkingCalculationOutputGraphQL.model_rebuild()
BenchmarkingCalculationOutputWrite.model_rebuild()
BenchmarkingCalculationOutputApply.model_rebuild()
BenchmarkingConfigurationDayAhead.model_rebuild()
BenchmarkingConfigurationDayAheadGraphQL.model_rebuild()
BenchmarkingConfigurationDayAheadWrite.model_rebuild()
BenchmarkingConfigurationDayAheadApply.model_rebuild()
BenchmarkingResultDayAhead.model_rebuild()
BenchmarkingResultDayAheadGraphQL.model_rebuild()
BenchmarkingResultDayAheadWrite.model_rebuild()
BenchmarkingResultDayAheadApply.model_rebuild()
BenchmarkingShopCase.model_rebuild()
BenchmarkingShopCaseGraphQL.model_rebuild()
BenchmarkingShopCaseWrite.model_rebuild()
BenchmarkingShopCaseApply.model_rebuild()
BenchmarkingTaskDispatcherInputDayAhead.model_rebuild()
BenchmarkingTaskDispatcherInputDayAheadGraphQL.model_rebuild()
BenchmarkingTaskDispatcherInputDayAheadWrite.model_rebuild()
BenchmarkingTaskDispatcherInputDayAheadApply.model_rebuild()
BenchmarkingTaskDispatcherOutputDayAhead.model_rebuild()
BenchmarkingTaskDispatcherOutputDayAheadGraphQL.model_rebuild()
BenchmarkingTaskDispatcherOutputDayAheadWrite.model_rebuild()
BenchmarkingTaskDispatcherOutputDayAheadApply.model_rebuild()
BidConfigurationDayAhead.model_rebuild()
BidConfigurationDayAheadGraphQL.model_rebuild()
BidConfigurationDayAheadWrite.model_rebuild()
BidConfigurationDayAheadApply.model_rebuild()
BidDocument.model_rebuild()
BidDocumentGraphQL.model_rebuild()
BidDocumentWrite.model_rebuild()
BidDocumentApply.model_rebuild()
BidDocumentAFRR.model_rebuild()
BidDocumentAFRRGraphQL.model_rebuild()
BidDocumentAFRRWrite.model_rebuild()
BidDocumentAFRRApply.model_rebuild()
BidDocumentDayAhead.model_rebuild()
BidDocumentDayAheadGraphQL.model_rebuild()
BidDocumentDayAheadWrite.model_rebuild()
BidDocumentDayAheadApply.model_rebuild()
BidMatrixInformation.model_rebuild()
BidMatrixInformationGraphQL.model_rebuild()
BidMatrixInformationWrite.model_rebuild()
BidMatrixInformationApply.model_rebuild()
BidRow.model_rebuild()
BidRowGraphQL.model_rebuild()
BidRowWrite.model_rebuild()
BidRowApply.model_rebuild()
FunctionOutput.model_rebuild()
FunctionOutputGraphQL.model_rebuild()
FunctionOutputWrite.model_rebuild()
FunctionOutputApply.model_rebuild()
Generator.model_rebuild()
GeneratorGraphQL.model_rebuild()
GeneratorWrite.model_rebuild()
GeneratorApply.model_rebuild()
MultiScenarioPartialBidMatrixCalculationInput.model_rebuild()
MultiScenarioPartialBidMatrixCalculationInputGraphQL.model_rebuild()
MultiScenarioPartialBidMatrixCalculationInputWrite.model_rebuild()
MultiScenarioPartialBidMatrixCalculationInputApply.model_rebuild()
PartialBidConfiguration.model_rebuild()
PartialBidConfigurationGraphQL.model_rebuild()
PartialBidConfigurationWrite.model_rebuild()
PartialBidConfigurationApply.model_rebuild()
PartialBidMatrixCalculationInput.model_rebuild()
PartialBidMatrixCalculationInputGraphQL.model_rebuild()
PartialBidMatrixCalculationInputWrite.model_rebuild()
PartialBidMatrixCalculationInputApply.model_rebuild()
PartialBidMatrixCalculationOutput.model_rebuild()
PartialBidMatrixCalculationOutputGraphQL.model_rebuild()
PartialBidMatrixCalculationOutputWrite.model_rebuild()
PartialBidMatrixCalculationOutputApply.model_rebuild()
PartialBidMatrixInformation.model_rebuild()
PartialBidMatrixInformationGraphQL.model_rebuild()
PartialBidMatrixInformationWrite.model_rebuild()
PartialBidMatrixInformationApply.model_rebuild()
PartialBidMatrixInformationWithScenarios.model_rebuild()
PartialBidMatrixInformationWithScenariosGraphQL.model_rebuild()
PartialBidMatrixInformationWithScenariosWrite.model_rebuild()
PartialBidMatrixInformationWithScenariosApply.model_rebuild()
PlantInformation.model_rebuild()
PlantInformationGraphQL.model_rebuild()
PlantInformationWrite.model_rebuild()
PlantInformationApply.model_rebuild()
PlantWaterValueBased.model_rebuild()
PlantWaterValueBasedGraphQL.model_rebuild()
PlantWaterValueBasedWrite.model_rebuild()
PlantWaterValueBasedApply.model_rebuild()
PriceAreaDayAhead.model_rebuild()
PriceAreaDayAheadGraphQL.model_rebuild()
PriceAreaDayAheadWrite.model_rebuild()
PriceAreaDayAheadApply.model_rebuild()
PriceAreaInformation.model_rebuild()
PriceAreaInformationGraphQL.model_rebuild()
PriceAreaInformationWrite.model_rebuild()
PriceAreaInformationApply.model_rebuild()
PriceProduction.model_rebuild()
PriceProductionGraphQL.model_rebuild()
PriceProductionWrite.model_rebuild()
PriceProductionApply.model_rebuild()
ShopBasedPartialBidConfiguration.model_rebuild()
ShopBasedPartialBidConfigurationGraphQL.model_rebuild()
ShopBasedPartialBidConfigurationWrite.model_rebuild()
ShopBasedPartialBidConfigurationApply.model_rebuild()
ShopCase.model_rebuild()
ShopCaseGraphQL.model_rebuild()
ShopCaseWrite.model_rebuild()
ShopCaseApply.model_rebuild()
ShopModel.model_rebuild()
ShopModelGraphQL.model_rebuild()
ShopModelWrite.model_rebuild()
ShopModelApply.model_rebuild()
ShopModelWithAssets.model_rebuild()
ShopModelWithAssetsGraphQL.model_rebuild()
ShopModelWithAssetsWrite.model_rebuild()
ShopModelWithAssetsApply.model_rebuild()
ShopPreprocessorInput.model_rebuild()
ShopPreprocessorInputGraphQL.model_rebuild()
ShopPreprocessorInputWrite.model_rebuild()
ShopPreprocessorInputApply.model_rebuild()
ShopPreprocessorOutput.model_rebuild()
ShopPreprocessorOutputGraphQL.model_rebuild()
ShopPreprocessorOutputWrite.model_rebuild()
ShopPreprocessorOutputApply.model_rebuild()
ShopResult.model_rebuild()
ShopResultGraphQL.model_rebuild()
ShopResultWrite.model_rebuild()
ShopResultApply.model_rebuild()
ShopScenario.model_rebuild()
ShopScenarioGraphQL.model_rebuild()
ShopScenarioWrite.model_rebuild()
ShopScenarioApply.model_rebuild()
ShopScenarioSet.model_rebuild()
ShopScenarioSetGraphQL.model_rebuild()
ShopScenarioSetWrite.model_rebuild()
ShopScenarioSetApply.model_rebuild()
ShopTriggerInput.model_rebuild()
ShopTriggerInputGraphQL.model_rebuild()
ShopTriggerInputWrite.model_rebuild()
ShopTriggerInputApply.model_rebuild()
ShopTriggerOutput.model_rebuild()
ShopTriggerOutputGraphQL.model_rebuild()
ShopTriggerOutputWrite.model_rebuild()
ShopTriggerOutputApply.model_rebuild()
TaskDispatcherInput.model_rebuild()
TaskDispatcherInputGraphQL.model_rebuild()
TaskDispatcherInputWrite.model_rebuild()
TaskDispatcherInputApply.model_rebuild()
TaskDispatcherOutput.model_rebuild()
TaskDispatcherOutputGraphQL.model_rebuild()
TaskDispatcherOutputWrite.model_rebuild()
TaskDispatcherOutputApply.model_rebuild()
TotalBidMatrixCalculationInput.model_rebuild()
TotalBidMatrixCalculationInputGraphQL.model_rebuild()
TotalBidMatrixCalculationInputWrite.model_rebuild()
TotalBidMatrixCalculationInputApply.model_rebuild()
TotalBidMatrixCalculationOutput.model_rebuild()
TotalBidMatrixCalculationOutputGraphQL.model_rebuild()
TotalBidMatrixCalculationOutputWrite.model_rebuild()
TotalBidMatrixCalculationOutputApply.model_rebuild()
WaterValueBasedPartialBidConfiguration.model_rebuild()
WaterValueBasedPartialBidConfigurationGraphQL.model_rebuild()
WaterValueBasedPartialBidConfigurationWrite.model_rebuild()
WaterValueBasedPartialBidConfigurationApply.model_rebuild()
WaterValueBasedPartialBidMatrixCalculationInput.model_rebuild()
WaterValueBasedPartialBidMatrixCalculationInputGraphQL.model_rebuild()
WaterValueBasedPartialBidMatrixCalculationInputWrite.model_rebuild()
WaterValueBasedPartialBidMatrixCalculationInputApply.model_rebuild()


__all__ = [
    "DataRecord",
    "DataRecordGraphQL",
    "DataRecordWrite",
    "ResourcesWrite",
    "DomainModel",
    "DomainModelCore",
    "DomainModelWrite",
    "DomainModelList",
    "DomainRelationWrite",
    "GraphQLCore",
    "GraphQLList",
    "ResourcesWriteResult",
    "PageInfo",
    "Alert",
    "AlertGraphQL",
    "AlertWrite",
    "AlertApply",
    "AlertList",
    "AlertWriteList",
    "AlertApplyList",
    "AlertFields",
    "AlertTextFields",
    "BenchmarkingCalculationInput",
    "BenchmarkingCalculationInputGraphQL",
    "BenchmarkingCalculationInputWrite",
    "BenchmarkingCalculationInputApply",
    "BenchmarkingCalculationInputList",
    "BenchmarkingCalculationInputWriteList",
    "BenchmarkingCalculationInputApplyList",
    "BenchmarkingCalculationInputFields",
    "BenchmarkingCalculationInputTextFields",
    "BenchmarkingCalculationOutput",
    "BenchmarkingCalculationOutputGraphQL",
    "BenchmarkingCalculationOutputWrite",
    "BenchmarkingCalculationOutputApply",
    "BenchmarkingCalculationOutputList",
    "BenchmarkingCalculationOutputWriteList",
    "BenchmarkingCalculationOutputApplyList",
    "BenchmarkingCalculationOutputFields",
    "BenchmarkingCalculationOutputTextFields",
    "BenchmarkingConfigurationDayAhead",
    "BenchmarkingConfigurationDayAheadGraphQL",
    "BenchmarkingConfigurationDayAheadWrite",
    "BenchmarkingConfigurationDayAheadApply",
    "BenchmarkingConfigurationDayAheadList",
    "BenchmarkingConfigurationDayAheadWriteList",
    "BenchmarkingConfigurationDayAheadApplyList",
    "BenchmarkingConfigurationDayAheadFields",
    "BenchmarkingConfigurationDayAheadTextFields",
    "BenchmarkingProductionObligationDayAhead",
    "BenchmarkingProductionObligationDayAheadGraphQL",
    "BenchmarkingProductionObligationDayAheadWrite",
    "BenchmarkingProductionObligationDayAheadApply",
    "BenchmarkingProductionObligationDayAheadList",
    "BenchmarkingProductionObligationDayAheadWriteList",
    "BenchmarkingProductionObligationDayAheadApplyList",
    "BenchmarkingProductionObligationDayAheadFields",
    "BenchmarkingProductionObligationDayAheadTextFields",
    "BenchmarkingResultDayAhead",
    "BenchmarkingResultDayAheadGraphQL",
    "BenchmarkingResultDayAheadWrite",
    "BenchmarkingResultDayAheadApply",
    "BenchmarkingResultDayAheadList",
    "BenchmarkingResultDayAheadWriteList",
    "BenchmarkingResultDayAheadApplyList",
    "BenchmarkingResultDayAheadFields",
    "BenchmarkingResultDayAheadTextFields",
    "BenchmarkingShopCase",
    "BenchmarkingShopCaseGraphQL",
    "BenchmarkingShopCaseWrite",
    "BenchmarkingShopCaseApply",
    "BenchmarkingShopCaseList",
    "BenchmarkingShopCaseWriteList",
    "BenchmarkingShopCaseApplyList",
    "BenchmarkingShopCaseFields",
    "BenchmarkingTaskDispatcherInputDayAhead",
    "BenchmarkingTaskDispatcherInputDayAheadGraphQL",
    "BenchmarkingTaskDispatcherInputDayAheadWrite",
    "BenchmarkingTaskDispatcherInputDayAheadApply",
    "BenchmarkingTaskDispatcherInputDayAheadList",
    "BenchmarkingTaskDispatcherInputDayAheadWriteList",
    "BenchmarkingTaskDispatcherInputDayAheadApplyList",
    "BenchmarkingTaskDispatcherInputDayAheadFields",
    "BenchmarkingTaskDispatcherInputDayAheadTextFields",
    "BenchmarkingTaskDispatcherOutputDayAhead",
    "BenchmarkingTaskDispatcherOutputDayAheadGraphQL",
    "BenchmarkingTaskDispatcherOutputDayAheadWrite",
    "BenchmarkingTaskDispatcherOutputDayAheadApply",
    "BenchmarkingTaskDispatcherOutputDayAheadList",
    "BenchmarkingTaskDispatcherOutputDayAheadWriteList",
    "BenchmarkingTaskDispatcherOutputDayAheadApplyList",
    "BenchmarkingTaskDispatcherOutputDayAheadFields",
    "BenchmarkingTaskDispatcherOutputDayAheadTextFields",
    "BidConfigurationDayAhead",
    "BidConfigurationDayAheadGraphQL",
    "BidConfigurationDayAheadWrite",
    "BidConfigurationDayAheadApply",
    "BidConfigurationDayAheadList",
    "BidConfigurationDayAheadWriteList",
    "BidConfigurationDayAheadApplyList",
    "BidConfigurationDayAheadFields",
    "BidConfigurationDayAheadTextFields",
    "BidDocument",
    "BidDocumentGraphQL",
    "BidDocumentWrite",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentWriteList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
    "BidDocumentAFRR",
    "BidDocumentAFRRGraphQL",
    "BidDocumentAFRRWrite",
    "BidDocumentAFRRApply",
    "BidDocumentAFRRList",
    "BidDocumentAFRRWriteList",
    "BidDocumentAFRRApplyList",
    "BidDocumentAFRRFields",
    "BidDocumentAFRRTextFields",
    "BidDocumentDayAhead",
    "BidDocumentDayAheadGraphQL",
    "BidDocumentDayAheadWrite",
    "BidDocumentDayAheadApply",
    "BidDocumentDayAheadList",
    "BidDocumentDayAheadWriteList",
    "BidDocumentDayAheadApplyList",
    "BidDocumentDayAheadFields",
    "BidDocumentDayAheadTextFields",
    "BidMatrix",
    "BidMatrixGraphQL",
    "BidMatrixWrite",
    "BidMatrixApply",
    "BidMatrixList",
    "BidMatrixWriteList",
    "BidMatrixApplyList",
    "BidMatrixFields",
    "BidMatrixTextFields",
    "BidMatrixInformation",
    "BidMatrixInformationGraphQL",
    "BidMatrixInformationWrite",
    "BidMatrixInformationApply",
    "BidMatrixInformationList",
    "BidMatrixInformationWriteList",
    "BidMatrixInformationApplyList",
    "BidMatrixInformationFields",
    "BidMatrixInformationTextFields",
    "BidRow",
    "BidRowGraphQL",
    "BidRowWrite",
    "BidRowApply",
    "BidRowList",
    "BidRowWriteList",
    "BidRowApplyList",
    "BidRowFields",
    "BidRowTextFields",
    "DateSpecification",
    "DateSpecificationGraphQL",
    "DateSpecificationWrite",
    "DateSpecificationApply",
    "DateSpecificationList",
    "DateSpecificationWriteList",
    "DateSpecificationApplyList",
    "DateSpecificationFields",
    "DateSpecificationTextFields",
    "FunctionInput",
    "FunctionInputGraphQL",
    "FunctionInputWrite",
    "FunctionInputApply",
    "FunctionInputList",
    "FunctionInputWriteList",
    "FunctionInputApplyList",
    "FunctionInputFields",
    "FunctionInputTextFields",
    "FunctionOutput",
    "FunctionOutputGraphQL",
    "FunctionOutputWrite",
    "FunctionOutputApply",
    "FunctionOutputList",
    "FunctionOutputWriteList",
    "FunctionOutputApplyList",
    "FunctionOutputFields",
    "FunctionOutputTextFields",
    "Generator",
    "GeneratorGraphQL",
    "GeneratorWrite",
    "GeneratorApply",
    "GeneratorList",
    "GeneratorWriteList",
    "GeneratorApplyList",
    "GeneratorFields",
    "GeneratorTextFields",
    "GeneratorEfficiencyCurve",
    "GeneratorEfficiencyCurveGraphQL",
    "GeneratorEfficiencyCurveWrite",
    "GeneratorEfficiencyCurveApply",
    "GeneratorEfficiencyCurveList",
    "GeneratorEfficiencyCurveWriteList",
    "GeneratorEfficiencyCurveApplyList",
    "GeneratorEfficiencyCurveFields",
    "MarketConfiguration",
    "MarketConfigurationGraphQL",
    "MarketConfigurationWrite",
    "MarketConfigurationApply",
    "MarketConfigurationList",
    "MarketConfigurationWriteList",
    "MarketConfigurationApplyList",
    "MarketConfigurationFields",
    "MarketConfigurationTextFields",
    "MultiScenarioPartialBidMatrixCalculationInput",
    "MultiScenarioPartialBidMatrixCalculationInputGraphQL",
    "MultiScenarioPartialBidMatrixCalculationInputWrite",
    "MultiScenarioPartialBidMatrixCalculationInputApply",
    "MultiScenarioPartialBidMatrixCalculationInputList",
    "MultiScenarioPartialBidMatrixCalculationInputWriteList",
    "MultiScenarioPartialBidMatrixCalculationInputApplyList",
    "MultiScenarioPartialBidMatrixCalculationInputFields",
    "MultiScenarioPartialBidMatrixCalculationInputTextFields",
    "PartialBidConfiguration",
    "PartialBidConfigurationGraphQL",
    "PartialBidConfigurationWrite",
    "PartialBidConfigurationApply",
    "PartialBidConfigurationList",
    "PartialBidConfigurationWriteList",
    "PartialBidConfigurationApplyList",
    "PartialBidConfigurationFields",
    "PartialBidConfigurationTextFields",
    "PartialBidMatrixCalculationInput",
    "PartialBidMatrixCalculationInputGraphQL",
    "PartialBidMatrixCalculationInputWrite",
    "PartialBidMatrixCalculationInputApply",
    "PartialBidMatrixCalculationInputList",
    "PartialBidMatrixCalculationInputWriteList",
    "PartialBidMatrixCalculationInputApplyList",
    "PartialBidMatrixCalculationInputFields",
    "PartialBidMatrixCalculationInputTextFields",
    "PartialBidMatrixCalculationOutput",
    "PartialBidMatrixCalculationOutputGraphQL",
    "PartialBidMatrixCalculationOutputWrite",
    "PartialBidMatrixCalculationOutputApply",
    "PartialBidMatrixCalculationOutputList",
    "PartialBidMatrixCalculationOutputWriteList",
    "PartialBidMatrixCalculationOutputApplyList",
    "PartialBidMatrixCalculationOutputFields",
    "PartialBidMatrixCalculationOutputTextFields",
    "PartialBidMatrixInformation",
    "PartialBidMatrixInformationGraphQL",
    "PartialBidMatrixInformationWrite",
    "PartialBidMatrixInformationApply",
    "PartialBidMatrixInformationList",
    "PartialBidMatrixInformationWriteList",
    "PartialBidMatrixInformationApplyList",
    "PartialBidMatrixInformationFields",
    "PartialBidMatrixInformationTextFields",
    "PartialBidMatrixInformationWithScenarios",
    "PartialBidMatrixInformationWithScenariosGraphQL",
    "PartialBidMatrixInformationWithScenariosWrite",
    "PartialBidMatrixInformationWithScenariosApply",
    "PartialBidMatrixInformationWithScenariosList",
    "PartialBidMatrixInformationWithScenariosWriteList",
    "PartialBidMatrixInformationWithScenariosApplyList",
    "PartialBidMatrixInformationWithScenariosFields",
    "PartialBidMatrixInformationWithScenariosTextFields",
    "Plant",
    "PlantGraphQL",
    "PlantWrite",
    "PlantApply",
    "PlantList",
    "PlantWriteList",
    "PlantApplyList",
    "PlantFields",
    "PlantTextFields",
    "PlantInformation",
    "PlantInformationGraphQL",
    "PlantInformationWrite",
    "PlantInformationApply",
    "PlantInformationList",
    "PlantInformationWriteList",
    "PlantInformationApplyList",
    "PlantInformationFields",
    "PlantInformationTextFields",
    "PlantWaterValueBased",
    "PlantWaterValueBasedGraphQL",
    "PlantWaterValueBasedWrite",
    "PlantWaterValueBasedApply",
    "PlantWaterValueBasedList",
    "PlantWaterValueBasedWriteList",
    "PlantWaterValueBasedApplyList",
    "PlantWaterValueBasedFields",
    "PlantWaterValueBasedTextFields",
    "PowerAsset",
    "PowerAssetGraphQL",
    "PowerAssetWrite",
    "PowerAssetApply",
    "PowerAssetList",
    "PowerAssetWriteList",
    "PowerAssetApplyList",
    "PowerAssetFields",
    "PowerAssetTextFields",
    "PriceArea",
    "PriceAreaGraphQL",
    "PriceAreaWrite",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "PriceAreaAFRR",
    "PriceAreaAFRRGraphQL",
    "PriceAreaAFRRWrite",
    "PriceAreaAFRRApply",
    "PriceAreaAFRRList",
    "PriceAreaAFRRWriteList",
    "PriceAreaAFRRApplyList",
    "PriceAreaAFRRFields",
    "PriceAreaAFRRTextFields",
    "PriceAreaDayAhead",
    "PriceAreaDayAheadGraphQL",
    "PriceAreaDayAheadWrite",
    "PriceAreaDayAheadApply",
    "PriceAreaDayAheadList",
    "PriceAreaDayAheadWriteList",
    "PriceAreaDayAheadApplyList",
    "PriceAreaDayAheadFields",
    "PriceAreaDayAheadTextFields",
    "PriceAreaInformation",
    "PriceAreaInformationGraphQL",
    "PriceAreaInformationWrite",
    "PriceAreaInformationApply",
    "PriceAreaInformationList",
    "PriceAreaInformationWriteList",
    "PriceAreaInformationApplyList",
    "PriceAreaInformationFields",
    "PriceAreaInformationTextFields",
    "PriceProduction",
    "PriceProductionGraphQL",
    "PriceProductionWrite",
    "PriceProductionApply",
    "PriceProductionList",
    "PriceProductionWriteList",
    "PriceProductionApplyList",
    "PriceProductionFields",
    "PriceProductionTextFields",
    "ShopAttributeMapping",
    "ShopAttributeMappingGraphQL",
    "ShopAttributeMappingWrite",
    "ShopAttributeMappingApply",
    "ShopAttributeMappingList",
    "ShopAttributeMappingWriteList",
    "ShopAttributeMappingApplyList",
    "ShopAttributeMappingFields",
    "ShopAttributeMappingTextFields",
    "ShopBasedPartialBidConfiguration",
    "ShopBasedPartialBidConfigurationGraphQL",
    "ShopBasedPartialBidConfigurationWrite",
    "ShopBasedPartialBidConfigurationApply",
    "ShopBasedPartialBidConfigurationList",
    "ShopBasedPartialBidConfigurationWriteList",
    "ShopBasedPartialBidConfigurationApplyList",
    "ShopBasedPartialBidConfigurationFields",
    "ShopBasedPartialBidConfigurationTextFields",
    "ShopCase",
    "ShopCaseGraphQL",
    "ShopCaseWrite",
    "ShopCaseApply",
    "ShopCaseList",
    "ShopCaseWriteList",
    "ShopCaseApplyList",
    "ShopCaseFields",
    "ShopCommands",
    "ShopCommandsGraphQL",
    "ShopCommandsWrite",
    "ShopCommandsApply",
    "ShopCommandsList",
    "ShopCommandsWriteList",
    "ShopCommandsApplyList",
    "ShopCommandsFields",
    "ShopCommandsTextFields",
    "ShopFile",
    "ShopFileGraphQL",
    "ShopFileWrite",
    "ShopFileApply",
    "ShopFileList",
    "ShopFileWriteList",
    "ShopFileApplyList",
    "ShopFileFields",
    "ShopFileTextFields",
    "ShopModel",
    "ShopModelGraphQL",
    "ShopModelWrite",
    "ShopModelApply",
    "ShopModelList",
    "ShopModelWriteList",
    "ShopModelApplyList",
    "ShopModelFields",
    "ShopModelTextFields",
    "ShopModelWithAssets",
    "ShopModelWithAssetsGraphQL",
    "ShopModelWithAssetsWrite",
    "ShopModelWithAssetsApply",
    "ShopModelWithAssetsList",
    "ShopModelWithAssetsWriteList",
    "ShopModelWithAssetsApplyList",
    "ShopOutputTimeSeriesDefinition",
    "ShopOutputTimeSeriesDefinitionGraphQL",
    "ShopOutputTimeSeriesDefinitionWrite",
    "ShopOutputTimeSeriesDefinitionApply",
    "ShopOutputTimeSeriesDefinitionList",
    "ShopOutputTimeSeriesDefinitionWriteList",
    "ShopOutputTimeSeriesDefinitionApplyList",
    "ShopOutputTimeSeriesDefinitionFields",
    "ShopOutputTimeSeriesDefinitionTextFields",
    "ShopPenaltyReport",
    "ShopPenaltyReportGraphQL",
    "ShopPenaltyReportWrite",
    "ShopPenaltyReportApply",
    "ShopPenaltyReportList",
    "ShopPenaltyReportWriteList",
    "ShopPenaltyReportApplyList",
    "ShopPenaltyReportFields",
    "ShopPenaltyReportTextFields",
    "ShopPreprocessorInput",
    "ShopPreprocessorInputGraphQL",
    "ShopPreprocessorInputWrite",
    "ShopPreprocessorInputApply",
    "ShopPreprocessorInputList",
    "ShopPreprocessorInputWriteList",
    "ShopPreprocessorInputApplyList",
    "ShopPreprocessorInputFields",
    "ShopPreprocessorInputTextFields",
    "ShopPreprocessorOutput",
    "ShopPreprocessorOutputGraphQL",
    "ShopPreprocessorOutputWrite",
    "ShopPreprocessorOutputApply",
    "ShopPreprocessorOutputList",
    "ShopPreprocessorOutputWriteList",
    "ShopPreprocessorOutputApplyList",
    "ShopPreprocessorOutputFields",
    "ShopPreprocessorOutputTextFields",
    "ShopResult",
    "ShopResultGraphQL",
    "ShopResultWrite",
    "ShopResultApply",
    "ShopResultList",
    "ShopResultWriteList",
    "ShopResultApplyList",
    "ShopResultFields",
    "ShopScenario",
    "ShopScenarioGraphQL",
    "ShopScenarioWrite",
    "ShopScenarioApply",
    "ShopScenarioList",
    "ShopScenarioWriteList",
    "ShopScenarioApplyList",
    "ShopScenarioFields",
    "ShopScenarioTextFields",
    "ShopScenarioSet",
    "ShopScenarioSetGraphQL",
    "ShopScenarioSetWrite",
    "ShopScenarioSetApply",
    "ShopScenarioSetList",
    "ShopScenarioSetWriteList",
    "ShopScenarioSetApplyList",
    "ShopScenarioSetFields",
    "ShopScenarioSetTextFields",
    "ShopTimeSeries",
    "ShopTimeSeriesGraphQL",
    "ShopTimeSeriesWrite",
    "ShopTimeSeriesApply",
    "ShopTimeSeriesList",
    "ShopTimeSeriesWriteList",
    "ShopTimeSeriesApplyList",
    "ShopTimeSeriesFields",
    "ShopTimeSeriesTextFields",
    "ShopTriggerInput",
    "ShopTriggerInputGraphQL",
    "ShopTriggerInputWrite",
    "ShopTriggerInputApply",
    "ShopTriggerInputList",
    "ShopTriggerInputWriteList",
    "ShopTriggerInputApplyList",
    "ShopTriggerInputFields",
    "ShopTriggerInputTextFields",
    "ShopTriggerOutput",
    "ShopTriggerOutputGraphQL",
    "ShopTriggerOutputWrite",
    "ShopTriggerOutputApply",
    "ShopTriggerOutputList",
    "ShopTriggerOutputWriteList",
    "ShopTriggerOutputApplyList",
    "ShopTriggerOutputFields",
    "ShopTriggerOutputTextFields",
    "TaskDispatcherInput",
    "TaskDispatcherInputGraphQL",
    "TaskDispatcherInputWrite",
    "TaskDispatcherInputApply",
    "TaskDispatcherInputList",
    "TaskDispatcherInputWriteList",
    "TaskDispatcherInputApplyList",
    "TaskDispatcherInputFields",
    "TaskDispatcherInputTextFields",
    "TaskDispatcherOutput",
    "TaskDispatcherOutputGraphQL",
    "TaskDispatcherOutputWrite",
    "TaskDispatcherOutputApply",
    "TaskDispatcherOutputList",
    "TaskDispatcherOutputWriteList",
    "TaskDispatcherOutputApplyList",
    "TaskDispatcherOutputFields",
    "TaskDispatcherOutputTextFields",
    "TotalBidMatrixCalculationInput",
    "TotalBidMatrixCalculationInputGraphQL",
    "TotalBidMatrixCalculationInputWrite",
    "TotalBidMatrixCalculationInputApply",
    "TotalBidMatrixCalculationInputList",
    "TotalBidMatrixCalculationInputWriteList",
    "TotalBidMatrixCalculationInputApplyList",
    "TotalBidMatrixCalculationInputFields",
    "TotalBidMatrixCalculationInputTextFields",
    "TotalBidMatrixCalculationOutput",
    "TotalBidMatrixCalculationOutputGraphQL",
    "TotalBidMatrixCalculationOutputWrite",
    "TotalBidMatrixCalculationOutputApply",
    "TotalBidMatrixCalculationOutputList",
    "TotalBidMatrixCalculationOutputWriteList",
    "TotalBidMatrixCalculationOutputApplyList",
    "TotalBidMatrixCalculationOutputFields",
    "TotalBidMatrixCalculationOutputTextFields",
    "TurbineEfficiencyCurve",
    "TurbineEfficiencyCurveGraphQL",
    "TurbineEfficiencyCurveWrite",
    "TurbineEfficiencyCurveApply",
    "TurbineEfficiencyCurveList",
    "TurbineEfficiencyCurveWriteList",
    "TurbineEfficiencyCurveApplyList",
    "TurbineEfficiencyCurveFields",
    "WaterValueBasedPartialBidConfiguration",
    "WaterValueBasedPartialBidConfigurationGraphQL",
    "WaterValueBasedPartialBidConfigurationWrite",
    "WaterValueBasedPartialBidConfigurationApply",
    "WaterValueBasedPartialBidConfigurationList",
    "WaterValueBasedPartialBidConfigurationWriteList",
    "WaterValueBasedPartialBidConfigurationApplyList",
    "WaterValueBasedPartialBidConfigurationFields",
    "WaterValueBasedPartialBidConfigurationTextFields",
    "WaterValueBasedPartialBidMatrixCalculationInput",
    "WaterValueBasedPartialBidMatrixCalculationInputGraphQL",
    "WaterValueBasedPartialBidMatrixCalculationInputWrite",
    "WaterValueBasedPartialBidMatrixCalculationInputApply",
    "WaterValueBasedPartialBidMatrixCalculationInputList",
    "WaterValueBasedPartialBidMatrixCalculationInputWriteList",
    "WaterValueBasedPartialBidMatrixCalculationInputApplyList",
    "WaterValueBasedPartialBidMatrixCalculationInputFields",
    "WaterValueBasedPartialBidMatrixCalculationInputTextFields",
    "Watercourse",
    "WatercourseGraphQL",
    "WatercourseWrite",
    "WatercourseApply",
    "WatercourseList",
    "WatercourseWriteList",
    "WatercourseApplyList",
    "WatercourseFields",
    "WatercourseTextFields",
]



from __future__ import annotations

import warnings
from typing import Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import field_validator, model_validator

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecord,
    DataRecordGraphQL,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    GraphQLCore,
    ResourcesWrite,
)


__all__ = [
    "GeneratorEfficiencyCurve",
    "GeneratorEfficiencyCurveWrite",
    "GeneratorEfficiencyCurveApply",
    "GeneratorEfficiencyCurveList",
    "GeneratorEfficiencyCurveWriteList",
    "GeneratorEfficiencyCurveApplyList",
    "GeneratorEfficiencyCurveFields",
]

GeneratorEfficiencyCurveFields = Literal["power", "efficiency"]

_GENERATOREFFICIENCYCURVE_PROPERTIES_BY_FIELD = {
    "power": "power",
    "efficiency": "efficiency",
}


class GeneratorEfficiencyCurveGraphQL(GraphQLCore):
    """This represents the reading version of generator efficiency curve, used
    when data is retrieved from CDF using GraphQL.

    It is used when retrieving data from CDF using GraphQL.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        data_record: The data record of the generator efficiency curve node.
        power: The generator power values
        efficiency: The generator efficiency values
    """

    view_id = dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1")
    power: Optional[list[float]] = None
    efficiency: Optional[list[float]] = None

    @model_validator(mode="before")
    def parse_data_record(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values
        if "lastUpdatedTime" in values or "createdTime" in values:
            values["dataRecord"] = DataRecordGraphQL(
                created_time=values.pop("createdTime", None),
                last_updated_time=values.pop("lastUpdatedTime", None),
            )
        return values

    def as_read(self) -> GeneratorEfficiencyCurve:
        """Convert this GraphQL format of generator efficiency curve to the reading format."""
        if self.data_record is None:
            raise ValueError("This object cannot be converted to a read format because it lacks a data record.")
        return GeneratorEfficiencyCurve(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecord(
                version=0,
                last_updated_time=self.data_record.last_updated_time,
                created_time=self.data_record.created_time,
            ),
            power=self.power,
            efficiency=self.efficiency,
        )

    def as_write(self) -> GeneratorEfficiencyCurveWrite:
        """Convert this GraphQL format of generator efficiency curve to the writing format."""
        return GeneratorEfficiencyCurveWrite(
            space=self.space or DEFAULT_INSTANCE_SPACE,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=0),
            power=self.power,
            efficiency=self.efficiency,
        )


class GeneratorEfficiencyCurve(DomainModel):
    """This represents the reading version of generator efficiency curve.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        data_record: The data record of the generator efficiency curve node.
        power: The generator power values
        efficiency: The generator efficiency values
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "GeneratorEfficiencyCurve"
    )
    power: Optional[list[float]] = None
    efficiency: Optional[list[float]] = None

    def as_write(self) -> GeneratorEfficiencyCurveWrite:
        """Convert this read version of generator efficiency curve to the writing version."""
        return GeneratorEfficiencyCurveWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            power=self.power,
            efficiency=self.efficiency,
        )

    def as_apply(self) -> GeneratorEfficiencyCurveWrite:
        """Convert this read version of generator efficiency curve to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class GeneratorEfficiencyCurveWrite(DomainModelWrite):
    """This represents the writing version of generator efficiency curve.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        data_record: The data record of the generator efficiency curve node.
        power: The generator power values
        efficiency: The generator efficiency values
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power_ops_types", "GeneratorEfficiencyCurve"
    )
    power: list[float]
    efficiency: list[float]

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
        allow_version_increase: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            GeneratorEfficiencyCurve, dm.ViewId("power_ops_core", "GeneratorEfficiencyCurve", "1")
        )

        properties: dict[str, Any] = {}

        if self.power is not None:
            properties["power"] = self.power

        if self.efficiency is not None:
            properties["efficiency"] = self.efficiency

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=None if allow_version_increase else self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        return resources


class GeneratorEfficiencyCurveApply(GeneratorEfficiencyCurveWrite):
    def __new__(cls, *args, **kwargs) -> GeneratorEfficiencyCurveApply:
        warnings.warn(
            "GeneratorEfficiencyCurveApply is deprecated and will be removed in v1.0. Use GeneratorEfficiencyCurveWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "GeneratorEfficiencyCurve.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class GeneratorEfficiencyCurveList(DomainModelList[GeneratorEfficiencyCurve]):
    """List of generator efficiency curves in the read version."""

    _INSTANCE = GeneratorEfficiencyCurve

    def as_write(self) -> GeneratorEfficiencyCurveWriteList:
        """Convert these read versions of generator efficiency curve to the writing versions."""
        return GeneratorEfficiencyCurveWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> GeneratorEfficiencyCurveWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class GeneratorEfficiencyCurveWriteList(DomainModelWriteList[GeneratorEfficiencyCurveWrite]):
    """List of generator efficiency curves in the writing version."""

    _INSTANCE = GeneratorEfficiencyCurveWrite


class GeneratorEfficiencyCurveApplyList(GeneratorEfficiencyCurveWriteList): ...


def _create_generator_efficiency_curve_filter(
    view_id: dm.ViewId,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None
